import json
import logging
import os
import subprocess
from typing import Optional, Generator, Any
import zipfile

import psutil
from diskcache import Cache
from dotenv import load_dotenv
from ga4gh.vrs import models as VRS
from ga4gh.vrs.dataproxy import _DataProxy, create_dataproxy
from ga4gh.vrs.extras.translator import AlleleTranslator
from pathlib import Path
from glom import glom
from pydantic import BaseModel, model_validator
import requests
import yaml

load_dotenv()

_logger = logging.getLogger(__name__)
LOGGED_ALREADY = set()
METAKB_API = "https://pediatric.metakb.org/api"


manifest: "Manifest" = None

# TODO - read from manifest
gigabytes = 20
bytes_in_a_gigabyte = 1024**3  # 1 gigabyte = 1024^3 bytes
cache_size_limit = gigabytes * bytes_in_a_gigabyte


def get_cache_directory(cache_dir: str, cache_name: str) -> str:
    """Return the cache directory."""
    return str(Path(cache_dir) / cache_name)


class CachingAlleleTranslator(AlleleTranslator):
    """A subclass of AlleleTranslator that uses cache results and adds a method to run in a threaded fashion."""

    _cache: Cache = None

    def __init__(self, data_proxy: _DataProxy, normalize: bool = False):
        super().__init__(data_proxy)
        self.normalize = normalize
        self._cache = None
        if manifest and manifest.cache_enabled:
            self._cache = Cache(
                directory=get_cache_directory(
                    manifest.cache_directory, "allele_translator"
                ),
                size_limit=cache_size_limit,
            )
        else:
            _logger.info("Cache is not enabled")

    def translate_from(self, var, fmt=None, **kwargs):
        """Check and update cache"""

        if self._cache is not None:
            key = f"{var}-{fmt}"
            if key in self._cache:
                return self._cache[key]

        allele = super().translate_from(var, fmt=fmt, **kwargs)

        assert isinstance(
            allele, VRS.Allele
        ), f"Allele is not the expected Pydantic Model {type(allele)}: {allele}"

        if self._cache is not None:
            self._cache[key] = allele.id

        return allele.id


def caching_allele_translator_factory(
    normalize: bool = True, seqrepo_uri: str | None = None
):
    """Return a CachingAlleleTranslator instance with a SeqRepo dataproxy."""
    if manifest is not None:
        dp = create_dataproxy(seqrepo_uri or manifest.seqrepo_uri)
    else:
        dp = create_dataproxy(seqrepo_uri)
    translator = CachingAlleleTranslator(dp)
    translator.normalize = normalize
    return translator


def generate_gnomad_ids(vcf_line, compute_for_ref: bool = True) -> list[str]:
    """Assuming a standard VCF format with tab-separated fields, generate a gnomAD-like ID from a VCF line.
    see https://github.com/ga4gh/vrs-python/blob/main/src/ga4gh/vrs/extras/vcf_annotation.py#L386-L411
    """
    fields = vcf_line.strip().split("\t")
    gnomad_ids = []
    # Extract relevant information (you may need to adjust these indices based on your VCF format)
    chromosome = fields[0]
    position = fields[1]
    reference_allele = fields[3]
    alternate_allele = fields[4]

    gnomad_loc = f"{chromosome}-{position}"
    if compute_for_ref:
        gnomad_ids.append(f"{gnomad_loc}-{reference_allele}-{reference_allele}")
    for alt in alternate_allele.split(","):
        alt = alt.strip()
        # TODO - Should we be raising a ValueError hear and let the caller do the logging?
        # TODO - Should this be a config in the manifest?
        # ['<INS>', '<DEL>', '<DUP>', '<INV>', '<CNV>', '<DUP:TANDEM>', '<DUP:INT>', '<DUP:EXT>', '*']
        invalid_alts = ["INS", "DEL", "DUP", "INV", "CNV", "TANDEM", "INT", "EXT", "*"]
        is_valid = True
        for invalid_alt in invalid_alts:
            if invalid_alt in alt:
                is_valid = False
                _ = f"Invalid alt found: {alt}"
                if _ not in LOGGED_ALREADY:
                    LOGGED_ALREADY.add(_)
                    _logger.error(_)
                break
        if is_valid:
            gnomad_ids.append(f"{gnomad_loc}-{reference_allele}-{alt}")

    return gnomad_ids


def params_from_vcf(path, limit=None) -> Generator[dict, None, None]:
    """Open the vcf file, skip headers, yield the first lines as gnomad-like IDs"""
    from vrs_anvil.translator import VCFItem

    c = 0
    with open(path, "r") as f:
        for line in f:
            if line.startswith("#"):
                continue
            gnomad_ids = generate_gnomad_ids(line)
            for gnomad_id in gnomad_ids:
                yield VCFItem(
                    fmt="gnomad",
                    var=gnomad_id,
                    file_name=path,
                    line_number=c,
                    identifier=None,
                )  # TODO - add identifier
            c += 1
            if limit and c > limit:
                break


def find_items_with_key(dictionary, key_to_find):
    """Find all items in a dictionary that have a specific key."""
    result = glom(dictionary, f"**.{key_to_find}")
    return result


class MetaKBProxy(BaseModel):
    """A proxy for the MetaKB, maintains a cache of VRS ids."""

    metakb_path: Path
    cache_path: Path
    _cache: Optional[Cache] = None

    def __init__(self, metakb_path: Path, cache_path: Path, cache: Cache = None):
        super().__init__(metakb_path=metakb_path, cache_path=cache_path, _cache=cache)
        if cache is None:
            reload_cache = False
            if not (metakb_path / "cache").is_dir():
                reload_cache = True
            cache = Cache(directory=get_cache_directory(cache_path, "metakb"))
            # cache.stats(enable=True) # drives up disk usage
            if reload_cache:
                for _ in metakb_ids(metakb_path):
                    cache.set(_, True)
        self._cache = cache

    def get(self, vrs_id: str) -> bool:
        """Get the vrs_id from the cache."""
        return self._cache.get(vrs_id, False)


def metakb_ids(metakb_path) -> Generator[str, None, None]:
    """Find all the applicable vrs ids in the metakb files."""
    if len(list(Path(metakb_path).glob("*.json"))) == 0:
        _get_metakb_models(metakb_path)

    for file_name in Path(metakb_path).glob("*.json"):
        if file_name.is_file():
            with open(file_name, "r") as file:
                data = json.loads(file.read())
                yield from (
                    [
                        _
                        for _ in find_items_with_key(data, "id")
                        if _.startswith("ga4gh:VA")
                    ]
                )


def _get_metakb_models(metakb_path):
    def _download_s3(url: str, outfile_path: Path) -> None:
        """Download objects from public s3 bucket

        :param url: URL for metakb file in s3 bucket
        :param outfile_path: Path where file should be saved
        """
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            with open(outfile_path, "wb") as h:
                for chunk in r.iter_content(chunk_size=8192):
                    if chunk:
                        h.write(chunk)

    Path(metakb_path).mkdir(exist_ok=True)

    date = "20240305"
    json_files = [f"civic_cdm_{date}.json", f"moa_cdm_{date}.json"]

    for json_file in json_files:
        json_path = f"{metakb_path}/{json_file}"

        url = (
            "https://vicc-metakb.s3.us-east-2.amazonaws.com"
            + f"/cdm/{date}/{json_file}.zip"
        )
        zip_path = f"{json_path}.zip"

        _download_s3(url, zip_path)
        with zipfile.ZipFile(zip_path, "r") as zip_ref:
            zip_ref.extractall(metakb_path)
        os.remove(zip_path)


class Manifest(BaseModel):
    """
    A class to represent the manifest file.
    Defaults to the values below if not provided in the manifest.yaml
    """

    cache_directory: str = "cache/"
    """Path to the cache directory, defaults to cache/ (relative to the root of the repository)"""

    num_threads: int = 2
    """Number of threads to use for processing, defaults to 2"""

    # TODO: not implemented
    annotate_vcfs: bool = False
    """Should we create new VCFs with annotations. FOR FUTURE USE"""

    state_directory: str = "state/"
    """where to store the state of the application, log files, etc."""

    vcf_files: list[str]
    """The local file paths or URLs to vcf files to be processed"""
    # TODO - 2x check why local files need to be absolute paths

    work_directory: str = "work/"
    """The directory to store intermediate files"""

    seqrepo_uri: str | None = None
    """Description of the available SeqRepo resource.

    Currently accepted URI schemes:

    * seqrepo+file:///path/to/seqrepo/root
    * seqrepo+:../relative/path/to/seqrepo/root
    * seqrepo+http://localhost:5000/seqrepo
    * seqrepo+https://somewhere:5000/seqrepo
    """

    normalize: bool = True
    """Normalize the VRS ids"""

    limit: Optional[int] = None
    """Stop processing after this many lines"""

    cache_enabled: Optional[bool] = True
    """Cache results"""

    compute_for_ref: Optional[bool] = False
    """Compute reference allele"""

    estimated_vcf_lines: Optional[int] = 4000000
    """How many lines per vcf file?  Used for progress bar"""

    metakb_directory: str = "metakb/"
    """Where the CDM files are located.  This is a directory containing json files"""

    disable_progress_bars: Optional[bool] = False

    @model_validator(mode="after")
    def check_paths(self) -> "Manifest":
        """Post init method to set the cache directory."""
        self.work_directory = str(Path(self.work_directory).expanduser())
        self.cache_directory = str(Path(self.cache_directory).expanduser())
        self.state_directory = str(Path(self.state_directory).expanduser())
        self.metakb_directory = str(Path(self.metakb_directory).expanduser())

        if not Path(self.metakb_directory).exists():
            raise ValueError("MetaKB directory does not exist")

        for _ in ["work_directory", "cache_directory", "state_directory"]:
            if not Path(getattr(self, _)).exists():
                Path(getattr(self, _)).mkdir(parents=True, exist_ok=True)
                _logger.debug(f"Created directory {getattr(self, _)}")

        return self


def query_metakb(
    variation_str: str | None = None,
    disease: str | None = None,
    therapy: str | None = None,
    gene: str | None = None,
    statement_id: str | None = None,
    log: bool = False,
) -> dict:
    """
    Query MetaKB API for variant statements.

    This is a more Pythonic wrapper around the MetaKB /search/statements endpoint.
    While the metakb package provides a direct Python API (search_statements), it requires
    a Neo4j database connection. Since we're querying the hosted API, we use HTTP requests
    but validate the response using the Pydantic models.

    Args:
        variation_str: Variation identifier (VRS ID, HGVS, gene symbol, etc.)
        disease: Optional disease filter
        therapy: Optional therapy filter
        gene: Optional gene filter
        statement_id: Optional statement ID filter (e.g., "civic.eid:102", "civic.aid:7")
        log: Whether to log the request (for debugging)

    Returns:
        Dictionary containing the search results (compatible with SearchStatementsResponse)

    Example:
        >>> # Query by freetext variation
        >>> results = query_metakb(variation_str="BRAF V600E")
        >>>
        >>> # Query by VRS ID
        >>> results = query_metakb(variation_str="ga4gh:VA.j4XnsLZcdzDIYa5pvvXM7t1wn9OITr0L")
        >>>
        >>> # Query by statement ID
        >>> results = query_metakb(statement_id="civic.eid:102")
        >>>
        >>> # Combine filters
        >>> results = query_metakb(variation_str="BRAF V600E", disease="melanoma")

    Raises:
        ValueError: If no query parameters are provided or if API returns an error
    """
    # Build query parameters
    params = {}
    if variation_str:
        params["variation"] = variation_str
    if disease:
        params["disease"] = disease
    if therapy:
        params["therapy"] = therapy
    if gene:
        params["gene"] = gene
    if statement_id:
        params["statement_id"] = statement_id

    # Validate that at least one parameter is provided
    if not params:
        raise ValueError(
            "At least one query parameter must be provided (variation, disease, therapy, gene, or statement_id)"
        )

    # Make the API request
    response = requests.get(
        f"{METAKB_API}/search/statements",
        params=params,
        headers={"Accept": "application/json"},
    )

    # Handle errors
    if response.status_code >= 400:
        error_msg = f"API error: {response.text} ({response.status_code})"
        _logger.error(error_msg)
        raise ValueError(error_msg)

    # Parse and return JSON response
    # The response can be validated with SearchStatementsResponse if needed:
    # validated = SearchStatementsResponse(limit=None, **response.json())
    return response.json()


def run_command_in_background(command) -> Any:
    """Execute the command in the background, return pid."""
    # Detach the process from the parent process (this process)
    if not isinstance(command, list):
        command = command.split()
    return subprocess.Popen(
        command, shell=False, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
    )


def get_process_info(pid):
    """Return the process information for the pid."""
    try:
        return psutil.Process(pid)
    except psutil.NoSuchProcess:
        return None


def save_manifest(manifest: Manifest, manifest_path: str):
    """pass in a Manifest and yaml path"""
    with open(manifest_path, "w") as stream:
        yaml.dump(manifest.model_dump(), stream)



import pathlib

import pytest

from vrs_anvil import metakb_ids, MetaKBProxy


@pytest.fixture
def expected_vrs_id_count():
    return 2986


@pytest.fixture
def expected_vrs_ids():
    return [
        "ga4gh:VA.SOEVGpU16hxYQtJNeRyfq0V-B0rSOGK-",
        "ga4gh:VA.fU8g-a8s0n-tFsj3-XbsuFe17MfySB4q",
        "ga4gh:VA.rRPCnh0XXjuePRGWerw6PhVXFYjhchwP",
    ]


def test_metakb_ids(
    metakb_directory, testing_manifest, expected_vrs_id_count, expected_vrs_ids
):
    """Test metakb ids."""
    vrs_ids = [vrs_id for vrs_id in metakb_ids(metakb_directory)]
    vrs_count = len(vrs_ids)
    assert (
        vrs_count >= expected_vrs_id_count
    ), f"Not enough VRS ids found in metakb {vrs_count} {vrs_ids}"

    metakb_proxy = MetaKBProxy(
        metakb_path=pathlib.Path(metakb_directory),
        cache_path=pathlib.Path(testing_manifest.cache_directory),
    )

    for id in vrs_ids:
        assert metakb_proxy.get(id), f"VRS id {id} not found in cache {id}"

    for id in expected_vrs_ids:
        assert metakb_proxy.get(id), f"Expected VRS id {id} not found in cache {id}"

    _, misses = metakb_proxy._cache.stats()
    assert misses == 0, f"Misses found in cache {misses}"



from plugin_system.plugins.base_plugin import BasePlugin
from plugin_system.utils import (
    load_dict,
    csv_to_dataframe,
    terra_data_table_to_dataframe,
)


class ThousandGenomesPlugin(BasePlugin):
    """
    Plugin for AnVIL 1000G PRIMED data release on Terra

    Link: https://anvil.terra.bio/#workspaces/anvil-datastorage/AnVIL_1000G_PRIMED-data-model

    Note that get_phenotype_index is inherited from the parent BasePlugin class.
    """

    def __init__(
        self, phenotype_table_path: str | None = None, index_path: str | None = None
    ):
        """constructor used to set a phenotype index if provided a file path for the index (index_path).
        Otherwise create a phenotype index using a Terra data table (no path specified) or with a csv/tsv filepath.

        Index example: {"sample_A": ["HP:0001263", "HP:0000002"], "sample_B": ["HP:0001263"]}
        Note that we actively do not use super() to invoke the BasePlugin's constructor to create custom functionality.

        Args:
            phenotype_table_path (str, optional): Path to csv/tsv of phenotype data specified by the GREGoR data model.
                When not specified, defaults to loading from Terra data table in existing workspace titled "phenotypes".
                For more info on the data model, see https://gregorconsortium.org/data-model. Defaults to None.
            index_path (str, optional): Path to existing phenotype index. Defaults to None.
        """

        self.phenotype_index = self.__create_phenotype_index(
            phenotype_table_path=phenotype_table_path, index_path=index_path
        )

    def __create_phenotype_index(
        self, phenotype_table_path: str | None = None, index_path: str | None = None
    ) -> dict[str, list[str]]:
        """[private method] given phenotypical data input specified by the GREGoR Data model (in either tsv/csv/Terra data table),
        return a dictionary mapping from each sample to its list of phenotypes

        Args:
            phenotype_table_path (str, optional): Path to csv/tsv of phenotype data specified by the GREGoR data model.
                    When not specified, defaults to loading from Terra data table in existing workspace titled "phenotypes".
                    For more info on the data model, see https://gregorconsortium.org/data-model
            index_path (str, optional): Path to pre-computed index. Defaults to None.

        Returns:
            dict[str, list[str]]: index of a sample id to sample's phenotypes.
        """

        # load index from file if already created
        if index_path is not None:
            return load_dict(index_path)

        # if no path specified, load phenotype table from Terra Data Table by default (must be in Terra workspace)
        if phenotype_table_path is None:
            phenotype_df = terra_data_table_to_dataframe(
                table_name="population_descriptor"
            )
        else:  # otherwise load phenotype data table from file
            phenotype_df = csv_to_dataframe(phenotype_table_path)

        # create participant to phenotypes mapping
        phenotype_index = {}
        for subject_id in phenotype_df["subject_id"].unique():
            all_phenotypes = phenotype_df[phenotype_df["subject_id"] == subject_id][
                "country_of_recruitment"
            ]

            phenotype_index[subject_id] = list(all_phenotypes.unique())

        return phenotype_index



import os
import pytest

from plugin_system.utils import terra_data_table_to_dataframe, WORKSPACE_ENV_KEYS


@pytest.fixture()
def table_name():
    return "phenotype"


@pytest.mark.skipif(
    any([os.getenv(key) is None for key in WORKSPACE_ENV_KEYS]),
    reason="can only be tested on Terra",
)
def test_terra_loads_gregor_phenotype_table_in_workspace(table_name):
    df = terra_data_table_to_dataframe(table_name)

    for field in ["participant_id", "presence", "term_id"]:
        assert (
            field in df.columns
        ), f"Could not find field named {field} in columns. \nAll columns: {df.columns}"



import time

import pytest
from unittest.mock import patch, MagicMock
from vrs_anvil.collector import collect_manifest_urls


@pytest.fixture
def mock_manifest():
    manifest = MagicMock()
    manifest.vcf_files = [
        "http://example.com/file1.vcf",
        "gs://bucket/file2.vcf",
        "s3://bucket/file3.vcf",
        "file://my/local/file4.vcf",
    ]
    manifest.num_threads = 20
    manifest.work_directory = "work/"
    return manifest


def download_google_blob_with_sleep(*args, **kwargs):
    time.sleep(1)
    return lambda bucket_name, source_blob_name, destination_file_name: "work/file2.vcf"


@patch("vrs_anvil.collector.download_http_file", return_value="work/file1.vcf")
@patch(
    target="vrs_anvil.collector.download_google_blob",
    new_callable=download_google_blob_with_sleep,
)
@patch("vrs_anvil.collector.download_s3_object", return_value="work/file3.vcf")
@patch(
    "vrs_anvil.collector.create_symlink_to_work_directory",
    return_value="work/file4.vcf",
)
def test_collect_manifest_urls(
    mock_download_s3_object,
    mock_download_google_blob,
    mock_download_http_file,
    mock_create_symlink_to_work_directory,
    mock_manifest,
):
    files = [_ for _ in collect_manifest_urls(mock_manifest)]
    print(files)
    # sort since the order of the files is not guaranteed
    assert sorted(files) == [
        "work/file1.vcf",
        "work/file2.vcf",
        "work/file3.vcf",
        "work/file4.vcf",
    ]



import pysam

from plugin_system.plugins.base_plugin import BasePlugin
from plugin_system.utils import (
    load_dict,
    csv_to_dataframe,
    terra_data_table_to_dataframe,
)


class GregorPlugin(BasePlugin):
    """
    Plugin for GREGoR U08 release data on Terra

    Note that get_phenotype_index is inherited from the parent BasePlugin class.
    """

    def __init__(
        self, phenotype_table_path: str | None = None, index_path: str | None = None
    ):
        """constructor used to set a phenotype index if provided a file path for the index (index_path).
        Otherwise create a phenotype index using a Terra data table (no path specified) or with a csv/tsv filepath.

        Index example: {"sample_A": ["HP:0001263", "HP:0000002"], "sample_B": ["HP:0001263"]}
        Note that we actively do not use super() to invoke the BasePlugin's constructor to create custom functionality.

        Args:
            phenotype_table_path (str, optional): Path to csv/tsv of phenotype data specified by the GREGoR data model.
                When not specified, defaults to loading from Terra data table in existing workspace titled "phenotypes".
                For more info on the data model, see https://gregorconsortium.org/data-model. Defaults to None.
            index_path (str, optional): Path to existing phenotype index. Defaults to None.
        """

        self.phenotype_index = self.__create_phenotype_index(
            phenotype_table_path=phenotype_table_path, index_path=index_path
        )

    def __create_phenotype_index(
        self, phenotype_table_path: str | None = None, index_path: str | None = None
    ) -> dict[str, list[str]]:
        """[private method] given phenotypical data input specified by the GREGoR Data model (in either tsv/csv/Terra data table),
        return a dictionary mapping from each sample to its list of phenotypes

        Args:
            phenotype_table_path (str, optional): Path to csv/tsv of phenotype data specified by the GREGoR data model.
                    When not specified, defaults to loading from Terra data table in existing workspace titled "phenotypes".
                    For more info on the data model, see https://gregorconsortium.org/data-model
            index_path (str, optional): Path to pre-computed index. Defaults to None.

        Returns:
            dict[str, list[str]]: index of a sample id to sample's phenotypes.
        """

        # load index from file if already created
        if index_path is not None:
            return load_dict(index_path)

        # if no path specified, load phenotype table from Terra Data Table by default (must be in Terra workspace)
        if phenotype_table_path is None:
            phenotype_df = terra_data_table_to_dataframe(table_name="phenotype")
        else:  # otherwise load phenotype data table from file
            phenotype_df = csv_to_dataframe(phenotype_table_path)

        # create participant to phenotypes mapping
        phenotype_index = {}
        for participant_id in phenotype_df["participant_id"].unique():
            all_phenotypes = phenotype_df[
                phenotype_df["participant_id"] == participant_id
            ]["term_id"]

            phenotype_index[participant_id] = list(all_phenotypes.unique())

        return phenotype_index

    def include_sample(
        self, sample_id: str, record: pysam.VariantRecord, phenotype: str
    ) -> bool:
        """determine whether to include a sample in the cohort allele frequency based on its variant data and phenotypic traits.
        Directly inherit implementation from base plugin

        Args:
            sample_id (str): sample_id used to uniquly identify a sample ID
            record (pysam.VariantRecord): PySam record object representing a VCF row
            phenotype (str): phenotype of interest, matching phenotype codes used in sample_phenotype_)index

        Returns:
            bool: whether to include the sample
        """

        # TODO: possibly implement filtering by sex of participant
        return super().include_sample(sample_id, record, phenotype)

    def process_sample_genotype(
        self,
        sample_id: str,
        record: pysam.VariantRecord,
        alt_index: int,
    ) -> tuple[int, int]:
        """given a sample's genotype, return focus and locus allele counts

        Args:
            sample_id (str): sample_id used to uniquly identify a sample ID
            record (pysam.VariantRecord): pysam record object representing a VCF row
            sample_phenotype_index (dict[str, list[str]]): mapping from sample IDs to each sample list of phenotypes
            alt_index (int): index matching the variant of interest

        Returns:
            tuple[int, int]: number of focus (specified) alleles, followed by number of locus (total) alleles.
        """

        # FIXME: support for hemizygous regions (chrY / mitochondrial variants)
        # GREGOR uses DRAGEN's continuous allele frequency approach:
        # https://support-docs.illumina.com/SW/DRAGEN_v40/Content/SW/DRAGEN/MitochondrialCalling.htm
        within_hemizygous_region = record.chrom in ["chrM", "chrY"]
        within_x_chr = record.chrom == "chrX"

        # get focus allele count, handling if there are multiple alts
        alleles = record.samples[sample_id].allele_indices
        num_focus_alleles = sum(
            [1 for _, alt_number in enumerate(alleles) if alt_number == alt_index]
        )

        if within_hemizygous_region and num_focus_alleles > 0:
            num_focus_alleles = 1

        # get total allele count
        if within_hemizygous_region:
            num_total_alleles = 1
        elif within_x_chr:
            # FIXME: make use of sex of participant?
            # by default considers all chrX samples as diploid (regardless of sex)
            is_female = True
            num_total_alleles = 2 if is_female else 1
        else:
            num_total_alleles = len(alleles)

        return num_focus_alleles, num_total_alleles



import logging
import queue
import threading
from dataclasses import dataclass, field
from typing import NamedTuple, Generator, Any, Optional

from pydantic import BaseModel

from vrs_anvil import caching_allele_translator_factory

_logger = logging.getLogger(__name__)


class WorkerThread(threading.Thread):
    """Read from the task queue, process the item, with local translator and write the result to the result queue."""

    def __init__(self, task_queue, result_queue, normalize):
        super().__init__(daemon=True)
        self.task_queue = task_queue
        self.result_queue = result_queue
        self.translator = caching_allele_translator_factory(normalize=normalize)
        self.busy = False

    def run(self):
        while True:
            try:
                prioritized_item = self.task_queue.get()
                item: VCFItem = prioritized_item.item
                if item is None:
                    break  # Signal to exit the thread

                self.busy = True
                allele_id = self.translator.translate_from(fmt=item.fmt, var=item.var)
                _ = item._asdict()
                _["result"] = allele_id
                result = VCFItem(**_)
                self.result_queue.put(PrioritizedItem(1, result))

                self.busy = False
                self.task_queue.task_done()

            except Exception as exc:
                _logger.exception(f"{self.name} error {exc}")
                self.busy = False
                break


class VCFItem(NamedTuple):
    """A named tuple to hold the VCF item."""

    fmt: str
    """The format of the item's id e.g gnomad - passed to vrs python"""
    var: str
    """variant - passed to vrs python"""
    file_name: str = None
    """data source"""
    line_number: int = None
    """line number in the file"""
    identifier: str = None
    """identifier for the item"""
    result: Any = None
    """identifier for the item"""


@dataclass(order=True)
class PrioritizedItem:
    """If the data elements are not comparable, the data can be wrapped in a class that ignores the data item and only compares the priority number"""

    priority: int
    item: Any = field(compare=False)


class Translator(BaseModel):
    """A class to run the translation in either threaded or non-threaded fashion."""

    normalize: Optional[bool] = False

    def translate_from(
        self, generator: Generator[VCFItem, None, None], num_threads: int = 8
    ) -> Generator[VCFItem, None, None]:
        if num_threads > 1:
            return threaded_translator(generator, num_threads, self.normalize)
        else:
            return inline_translator(generator, self.normalize)


def inline_translator(
    generator: Generator[VCFItem, None, None], normalize: bool = False
) -> Generator[VCFItem, None, None]:
    """A generator that runs the translation in a non-threaded fashion."""
    tlr = caching_allele_translator_factory(normalize=normalize)
    for item in generator:
        allele_id = tlr.translate_from(fmt=item.fmt, var=item.var)
        _ = item._asdict()
        _["result"] = allele_id
        yield VCFItem(**_)


def threaded_translator(
    generator: Generator[VCFItem, None, None],
    num_worker_threads: int,
    normalize: bool = False,
) -> Generator[VCFItem, None, None]:
    """A generator that runs the translation in a threaded fashion."""
    task_queue = queue.PriorityQueue(maxsize=num_worker_threads * 2)
    result_queue = queue.PriorityQueue(maxsize=num_worker_threads * 2)

    # Start worker threads
    worker_threads = [
        WorkerThread(task_queue, result_queue, normalize)
        for _ in range(num_worker_threads)
    ]
    for worker_thread in worker_threads:
        worker_thread.start()

    # Start the reader thread

    def reader_thread(_generator):
        for item in _generator:
            task_queue.put(PrioritizedItem(1, item))  # Default priority is 1

    reader = threading.Thread(target=reader_thread, args=(generator,), daemon=True)
    reader.start()

    # Main thread yields results
    logged_already = []
    c = 0
    while True:
        try:
            prioritized_item = result_queue.get(
                timeout=1
            )  # Give the thread time to start
            c += 1
            yield prioritized_item.item
            result_queue.task_done()
        except queue.Empty:
            reader_started = c > 0
            if reader_started and all(
                not worker_thread.busy for worker_thread in worker_threads
            ):
                _logger.info("All worker threads are idle, exiting.")
                break
            if not reader_started:
                msg = "queue.Empty - Waiting for reader to start"
            else:
                msg = "queue.Empty - Waiting for worker threads to finish"

            if msg not in logged_already:
                _logger.info(msg)
                logged_already.append(msg)



"""
Get statistics from 1000g samples, including
  - total samples
  - % matches to each kb individually and metakb as a whole
  - study ids and descriptions by variant hit in metakb

caveats: variant match on vrs id, not on
  - matching zygosity
  - germline vs somatic variant

to run...
  - create a tmp/ directory in the repo root directory
  - download the chr1 vcf.gz file from Terra to
    work/1kGP_high_coverage_Illumina.chr1.filtered.SNV_INDEL_SV_phased_panel.vcf.gz
  - cd tmp/
  - python ../scripts/1000g-sample-stats.py
"""

import json
import pathlib
import matplotlib.pyplot as plt
import os
import numpy as np
import pandas as pd
import pysam
import seaborn as sns
import yaml

from collections import defaultdict
from glob import glob
from vrs_anvil import query_metakb
from vrs_anvil.annotator import MATCHES, TOTAL, VRS_OBJECT

# grabs all metrics files from this directory
metrics_dir = "../tests/fixtures/chr1_metrics"

# settings and files to load
figure_dir = "figures"
variant_percentages_file_name = "chr1.png"
variant_histogram_file_name = "chr1_variants_per_patient"
save_figures = False
show_figures = True

# cohort allele frequency info
caf_dicts = []

# seaborn styling
sns.set_theme()
sns.set_style("whitegrid")


# helper functions
def truncate(s, first_few, last_few):
    "truncate string printing only first_few and last_few characters"
    return f"{s[:first_few]}...{s[-last_few:]}"


def create_caf_dict(
    allele_id,
    gnomad_expression,
    focus_allele,
    focus_allele_count,
    locus_allele_count,
    ancillary_results=None,
):
    allele_frequency = focus_allele_count * 1.0 / locus_allele_count

    caf_dict = {
        "id": allele_id,
        "type": "CohortAlleleFrequency",
        "label": f"Overall Cohort Allele Frequency for {allele_id} ({gnomad_expression})",
        "derivedFrom": {
            "id": f"AnVIL_1000G_PRIMED-data-model",
            "type": "DataSet",
            "label": f"AnVIL 1000G PRIMED Data Model",
            "version": "Last Updated 3/12/2024",
        },
        "focusAllele": focus_allele,
        "focusAlleleCount": focus_allele_count,
        "locusAlleleCount": locus_allele_count,
        "alleleFrequency": allele_frequency,
        "cohort": {
            "id": "all_3202_samples",
            "label": "High-coverage sequencing of 3202 samples",
        },
    }

    if ancillary_results is not None:
        caf_dict["ancillaryResults"] = metakb_response

    return caf_dict


# data storage
sample_evidence_dict = defaultdict(list)
variant_types_set = set()
matches_per_file = {}

# get evidence key drill down for each ting
METAKB_DIR = f"../tests/fixtures/metakb"
json_paths = list(pathlib.Path(METAKB_DIR).glob("*.json"))

# collect all evidence from a list of metrics files in a directory
all_metrics_paths = glob(f"{metrics_dir}/metrics*.yaml")

for metrics_path in all_metrics_paths:
    with open(metrics_path, "r") as file:
        metrics = yaml.safe_load(file)

    for file_path in metrics:
        if file_path == TOTAL:
            continue

        # load original vcf
        original_path = os.path.realpath(file_path)
        vcf_reader = pysam.VariantFile(original_path)
        num_samples = len(vcf_reader.header.samples)
        assert (
            num_samples == 3202
        ), f"expected 3202 samples for {file_path}, got {num_samples}"

        # check if matches to metakb cache found
        print("from file", truncate(original_path, 0, 46))
        if MATCHES in metrics[file_path]:
            print(f"{len(metrics[file_path][MATCHES])} hits found")
            matches_per_file.update({original_path: metrics[file_path][MATCHES]})
        else:
            print("no evidence from this file")
        print()

# get study info (id: description) for each variant
for file_path, evidence in matches_per_file.items():
    vcf_reader = pysam.VariantFile(file_path)

    print(truncate(file_path, 0, 47))

    for allele_id, allele_info in evidence.items():

        # extract coordinate information
        gnomad_expr = allele_info["parameters"]["var"]
        chrom, pos, ref, alt = gnomad_expr.split("-")
        pos = int(pos)
        print(f"\tgnomad: {gnomad_expr}")
        print(f"\tallele_id: {allele_id}")

        # get study id associated with vrs allele id
        metakb_response = query_metakb(allele_id)
        if metakb_response is None:
            print(f"no metakb hit for allele {allele_id}\n")
        else:
            study_ids = metakb_response["study_ids"]

            assert study_ids == [
                study["id"] for study in metakb_response["studies"]
            ], "study ids in wrong order to add variant types"
            variant_types = [
                study["qualifiers"]["alleleOrigin"]
                for study in metakb_response["studies"]
            ]
            variant_types_set.update(variant_types)

            for study in metakb_response["studies"]:
                print(f"\t{study['type']} ({study['id']}): {study['description']}")

            # should be a single record each time
            if sum([1 for _ in vcf_reader.fetch(chrom, pos - 1, pos)]) != 1:
                print(f"[WARNING] more than one record found at pos {pos}")

            for record_idx, record in enumerate(vcf_reader.fetch(chrom, pos - 1, pos)):
                print(
                    f"scanning row {record.chrom}, {record.pos}, {record.ref}, {record.alts}"
                )

                assert ref == record.ref, f"expected ref {record.ref}, got {ref}"
                if alt not in record.alts:
                    print(
                        f"Skipping {gnomad_expr}, expected one of {record.alts} alts, got {alt}"
                    )
                    continue

                # populate patient/sample dictionary
                id_count = 0

                for sample, genotype in record.samples.items():
                    gt_values = genotype["GT"]

                    # Check if any allele is non-zero
                    if any(int(allele) > 0 for allele in gt_values):
                        id_count += 1
                        if sample not in sample_evidence_dict:
                            sample_evidence_dict[sample] = {}

                        if "vrs_ids" in sample_evidence_dict[sample]:
                            sample_evidence_dict[sample]["vrs_ids"].append(allele_id)
                            sample_evidence_dict[sample]["study_ids"].extend(study_ids)
                            sample_evidence_dict[sample]["variant_types"].extend(
                                variant_types
                            )
                        else:
                            sample_evidence_dict[sample]["vrs_ids"] = [allele_id]
                            sample_evidence_dict[sample]["study_ids"] = list(study_ids)
                            sample_evidence_dict[sample]["variant_types"] = list(
                                variant_types
                            )

                print(f"\ttotal count: {id_count}\n")

        # get cohort allele counts
        for record_idx, record in enumerate(vcf_reader.fetch(chrom, pos - 1, pos)):
            print(
                f"[CAF] summing row {record.chrom}, {record.pos}, {record.ref}, {record.alts}"
            )

            # check correct row
            assert ref == record.ref, f"expected ref {record.ref}, got {ref}"
            if alt not in record.alts:
                print(
                    f"Skipping {gnomad_expr}, expected one of {record.alts} alts, got {alt}"
                )
                continue

            # get number of alleles across all patients
            focus_allele_count = sum(
                [sum(genotype["GT"]) for _, genotype in record.samples.items()]
            )
            print(focus_allele_count)

        # create caf dict for each metakb cache hit
        locus_allele_count = num_samples * 2
        allele_freq = focus_allele_count * 1.0 / (num_samples * 2)

        caf_dict = create_caf_dict(
            allele_id,
            gnomad_expr,
            allele_info[VRS_OBJECT],
            focus_allele_count,
            locus_allele_count,
            metakb_response,
        )
        print(f"Adding {caf_dict['label']}\n")
        caf_dicts.append(caf_dict)


#### Figures ####
# patients with one match total
def get_percent(a, b, output=True):
    "pretty print percentages"
    percent = float(f"{(100.0*a/b):.1f}")
    if output:
        print(f"{a}/{b} = {percent}%")
    return percent


print("patients with at least one variant match:", end=" ")
num_samples = len(vcf_reader.header.samples)

knowledgebases = ["MOAlmanac", "CIVIC", "All Knowledgebases"]
kb_keywords = ["moa", "civic", ""]  # "" represents match for everything

KB, PCT, VAR = range(3)
cols = ["knowledgebase", "percent", "variant_type"]
dtypes = [str, float, str]

TOTAL = "all"
variants = sorted(list(variant_types_set)) + [TOTAL]
data = None

# for each variant type (germline, somatic) and kb, get count of people in
for variant_type in variants:
    for i, keyword in enumerate(kb_keywords):
        num_matching_samples = 0

        # increment if sample has matching variant and kb
        for _, id_lists in sample_evidence_dict.items():
            v_types = np.array(id_lists["variant_types"])

            if variant_type == TOTAL:
                study_ids = np.array(id_lists["study_ids"])
            else:
                study_ids = np.array(id_lists["study_ids"])[v_types == variant_type]
            if any(keyword in study_id for study_id in study_ids):
                num_matching_samples += 1

        percent = num_matching_samples * 100.0 / num_samples
        if data is None:
            data = np.array([knowledgebases[i], percent, variant_type])
        else:
            data = np.vstack([data, [knowledgebases[i], percent, variant_type]])

expected_num_rows = len(knowledgebases) * len(variants)
assert (
    len(data) == expected_num_rows
), f"expected {expected_num_rows} rows, got {len(data)}"

# bar plot
df_pct = pd.DataFrame(data, columns=cols).astype(
    dtype={col: dtype for col, dtype in zip(cols, dtypes)}
)
ax = sns.barplot(
    data=df_pct,
    x=cols[VAR],
    y=cols[PCT],
    hue=cols[KB],
    order=["somatic", "germline", TOTAL],
)


# add percent labels
def add_percent_labels(num_bars, ax, vertical_nudge, from_decimal=False):
    multiplier = 100 if from_decimal else 1

    for i, bar in enumerate(ax.patches):
        if i >= num_bars:
            break

        height = bar.get_height()
        percent = f"{multiplier * height:.2f}"
        x = bar.get_x() + bar.get_width() / 2
        y = bar.get_y() + height + vertical_nudge
        ax.text(x, y, percent, ha="center", va="center", fontsize=10)


add_percent_labels(len(df_pct), ax, 2.5)

# sns.despine(left=True)
plt.ylim(0, 100)
plt.yticks(np.arange(0, 101, 20))

# Set labels and title
plt.xlabel("Knowledgebases")
plt.ylabel("Percent of Patients with Match (%)")
plt.title(
    "[chr1 to chr2] Percent of 1000G Patients with a Variant Match (n=3202)",
    wrap=True,
)
plt.legend(loc="upper left")

# Show the plot
plt.tight_layout()
os.makedirs(figure_dir, exist_ok=True)
if save_figures:
    plt.savefig(f"{figure_dir}/{variant_percentages_file_name}", dpi=300)
if show_figures:
    plt.show()

# average number of variants for all samples
num_variant_hits = sum(len(e) for e in matches_per_file.values())
print("total hits across all patients:", num_variant_hits)

jitter = [0.008, 0.018, 0.004]


NUM_VARIANTS = "num_variants"

for i, variant_type in enumerate(variants):
    num_variants_per_patient = [
        len(
            [
                v
                for v in values["variant_types"]
                if v == variant_type or variant_type == TOTAL
            ]
        )
        for values in sample_evidence_dict.values()
    ]

    num_variants_per_patient.extend(
        [0 for _ in range(num_samples - len(sample_evidence_dict))]
    )
    assert (
        len(num_variants_per_patient) == num_samples
    ), f"only {len(num_variants_per_patient)} samples, expected {num_samples}"

    df = pd.DataFrame({NUM_VARIANTS: num_variants_per_patient})
    df["percentage"] = df[NUM_VARIANTS].value_counts(normalize=True) * 100

    plt.figure()
    ax = sns.histplot(data=df, x=NUM_VARIANTS, stat="density", discrete=True)
    plt.xlabel("Number of Variants")
    plt.ylabel("Percentage of All Patients (%)")
    plt.title(
        f"Number of {variant_type.capitalize()} Variants Associated with Each Patient"
    )

    variant_hits_by_type = max((df[NUM_VARIANTS]) + 1)
    add_percent_labels(variant_hits_by_type, ax, jitter[i], from_decimal=True)

    plt.gca().yaxis.set_major_formatter(lambda x, _: f"{(x*100):.0f}")
    plt.grid(False)
    plt.xticks(range(min(df[NUM_VARIANTS]), variant_hits_by_type))

    if save_figures:
        plt.savefig(f"{figure_dir}/{variant_histogram_file_name}", dpi=300)
    if show_figures:
        plt.show()



import os
from pathlib import Path
import shutil
import pytest

from click.testing import CliRunner
from glob import glob
from unittest.mock import MagicMock, patch
from vrs_anvil import Manifest
from vrs_anvil.cli import cli

############
# FIXTURES #
############


@pytest.fixture
def num_vcfs(testing_manifest):
    return len(testing_manifest.vcf_files)


@pytest.fixture
def recent_timestamp():
    return "20240507_203113"


@pytest.fixture
def ps_dir():
    return "tests/fixtures/ps"


@pytest.fixture
def ps_manifest_path(ps_dir):
    return ps_dir + "/manifest.yaml"


@pytest.fixture
def mock_cli_manifest(tmp_path, monkeypatch, manifest_path, testing_manifest):
    """Move relevant files and working dir into a temporary dir before testing cli commands"""

    # copy manifest and vcf files to tmp_path
    shutil.copy(manifest_path, tmp_path)
    print(tmp_path)

    for file_path in testing_manifest.vcf_files:
        path_arr = str(Path(file_path)).split("/")
        if len(path_arr) > 1:
            file_dir = "/".join(path_arr[:-1])
            os.makedirs(f"{tmp_path}/{file_dir}", exist_ok=True)
        shutil.copy(file_path, f"{tmp_path}/{file_path}")

    # create metakb cdm path
    metakb_rel_path = str(Path(testing_manifest.metakb_directory))
    metakb_test_dir = f"{tmp_path}/{str(metakb_rel_path)}"
    os.makedirs(metakb_test_dir)
    shutil.copytree(metakb_rel_path, metakb_test_dir, dirs_exist_ok=True)

    # change to temp testing directory
    monkeypatch.chdir(tmp_path)
    return testing_manifest


#########
# TESTS #
#########


@pytest.fixture
def suffix():
    return "arbitrary_suffix"


def test_cli_version():
    """Test that version can be called and printed"""
    runner = CliRunner()
    result = runner.invoke(cli, "--version")
    expected_strings = ["version"]
    print(result.output)
    for expected_string in expected_strings:
        assert (
            expected_string in result.output
        ), f"Should have printed {expected_string}"


def test_loading_manifest(mock_cli_manifest):
    # mock_cli_manifest used to set up directories

    """Test that manifest can be parsed and stored in context"""
    runner = CliRunner()
    result = runner.invoke(cli, "--verbose")
    expected_strings = Manifest.model_fields.keys()
    print(result.output)
    for expected_string in expected_strings:
        assert (
            expected_string in result.output
        ), f"Should have printed {expected_string}"


def test_using_suffix(mock_cli_manifest, suffix):
    # mock_cli_manifest used to set up directories
    """Test that the suffix is successfully added to the manifest string"""

    # stub the annotate_all function
    # note the path is to the namespace (module) being tested not the namespace its imported from
    mock = MagicMock()
    mock.return_value = "mocked_metrics.yaml"
    runner = CliRunner()
    with patch("vrs_anvil.cli.annotate_all", mock):
        result = runner.invoke(cli, f"--suffix {suffix} annotate")

    print(result.output)
    assert suffix in result.output, f"Should have printed {suffix}"


def test_annotate_scatter(mock_cli_manifest):
    """Test that the annotate command scatter saves the right files"""

    # setup
    manifest = mock_cli_manifest
    print("manifest", manifest)
    runner = CliRunner()

    # stub the annotate_all function
    mock = MagicMock()

    # run annotate scatter cmd
    with patch("vrs_anvil.cli.run_command_in_background", mock):
        result = runner.invoke(cli, "annotate --scatter")
    print(result.output)

    # make sure scattered processes and log files
    work_dir = str(Path(manifest.work_directory))
    process_files = len(glob(f"{work_dir}/*scattered_process*.yaml"))
    assert process_files == 1, f"expected 1 scattered process file, got {process_files}"

    state_dir = str(Path(manifest.state_directory))
    num_log_files = len(glob(f"{state_dir}/*.log"))
    assert num_log_files == 1, f"expected 1 log files, got {num_log_files}"


def test_ps_returns_recent_files(ps_dir, monkeypatch, recent_timestamp, num_vcfs):
    """Test that vrs_anvil ps returns the most recent scatter command"""

    # make function call
    runner = CliRunner()
    monkeypatch.chdir(ps_dir)

    result = runner.invoke(cli, "--manifest manifest.yaml ps")
    print(result.output)

    # check successful command and loaded most recent process file
    assert result.exit_code == 0, f"result failed with message: \n{result}"
    assert (
        "no scattered processes" not in result.output
    ), "no scattered processes located"
    assert recent_timestamp in result.output, "most recent date has not been chosen"

    # check metrics and manifest files located
    for i in range(num_vcfs):
        assert (
            f"manifest_scattered_{recent_timestamp}_{i}.yaml" in result.output
        ), f"manifest #{i} of {num_vcfs} not found"

        assert (
            f"metrics_scattered_{recent_timestamp}_{i}.yaml" in result.output
        ), f"metrics file #{i} of {num_vcfs} not found"



import logging
import pathlib

import pytest
import yaml

from vrs_anvil.translator import Translator
from vrs_anvil import caching_allele_translator_factory, Manifest

_logger = logging.getLogger(__name__)
_logger.setLevel(logging.DEBUG)


@pytest.fixture
def manifest_path() -> pathlib.Path:
    """Return a path to a manifest file."""
    path = pathlib.Path("tests/fixtures/manifest.yaml").resolve()
    assert path.exists()
    return path


# note that tmp_path is a fixture provided by pytest
@pytest.fixture
def testing_manifest(manifest_path: pathlib.Path, tmp_path) -> Manifest:
    with open(manifest_path, "r") as stream:
        manifest_dict = yaml.safe_load(stream)

        # use pytest-provided relative path
        # note that metakb_directory is omitted to refer to original cdms
        for dir_key in ["work_directory", "cache_directory", "state_directory"]:
            manifest_dict[dir_key] = str(tmp_path / manifest_dict[dir_key])

        manifest = Manifest.model_validate(manifest_dict)
    return manifest


@pytest.fixture
def metakb_directory(testing_manifest):
    """Where the CDM files are located."""
    return testing_manifest.metakb_directory


@pytest.fixture
def seqrepo_uri(testing_manifest: Manifest):
    """Return the seqrepo directory as fixture."""
    return testing_manifest.seqrepo_uri


@pytest.fixture
def caching_translator(testing_manifest: Manifest):
    """Return a single translator instance."""
    return caching_allele_translator_factory(
        seqrepo_uri=testing_manifest.seqrepo_uri
    )


@pytest.fixture
def translator():
    """Return a translator instance."""
    return Translator(normalize=False)


@pytest.fixture()
def num_threads(testing_manifest):
    """Return the number of threads to use for testing."""
    return testing_manifest.num_threads


@pytest.fixture
def thousand_genome_vcf() -> pathlib.Path:
    """Return a path to a 1000g vcf."""
    _ = pathlib.Path("tests/fixtures/1kGP.chr1.1000.vcf")
    assert _.exists()
    return _


@pytest.fixture
def python_source_directories() -> list[str]:
    """Directories to scan with flake8."""
    return ["src", "tests"]



from ga4gh.vrs.extras.translator import AlleleTranslator
from ga4gh.vrs.extras.annotator.vcf import VcfAnnotator
from vrs_anvil import seqrepo_dir

SEQREPO_DIR = seqrepo_dir()
input_vcf = "tests/fixtures/1kGP.chr1.1000.vcf"
input_vcf_path_stem = input_vcf.replace(".vcf", "")
output_pkl = f"{input_vcf_path_stem}-test-vrs-objects.pkl"

output_vcf = None
require_validation = True
rle_seq_limit = 50

vcf_annotator = VcfAnnotator(seqrepo_root_dir=SEQREPO_DIR)
vcf_annotator.tlr.rle_seq_limit = rle_seq_limit
vcf_annotator.annotate(
    vcf_in=input_vcf,
    vcf_out=output_vcf,
    vrs_pickle_out=output_pkl,
    require_validation=require_validation,
)

print(f"written to {output_pkl}")



import gzip
import logging
import pathlib
import time
from collections import defaultdict
from datetime import datetime
from typing import Generator

import yaml
from ga4gh.vrs import models as VRS
from tqdm import tqdm

import vrs_anvil
from vrs_anvil import Manifest, generate_gnomad_ids
from vrs_anvil.collector import collect_manifest_urls
from vrs_anvil.translator import Translator, VCFItem

_logger = logging.getLogger(__name__)

# enums for metrics
# TODO: do this for keys across files like "parameters" but also "fmt" and "line"
PARAMETERS = "parameters"
TOTAL = "total"
STATUS = "status"
SUCCESSES = "successes"
ERROR = "error"
METAKB_HITS = "metakb_hits"
MATCHES = "matches"
START_TIME = "start_time"
END_TIME = "end_time"
ELAPSED_TIME = "elapsed_time"
LINE_COUNT = "line_count"
VRS_OBJECT = "vrs_object"
TIMESTAMP = "timestamp_str"


def recursive_defaultdict():
    """Implicitly create an entry if a key is read that doesn’t yet exist, any level deep"""
    return defaultdict(recursive_defaultdict)


metrics = recursive_defaultdict()


def _work_file_generator(manifest: Manifest) -> Generator[pathlib.Path, None, None]:
    """Return a generator for the files in the manifest."""
    for work_file in collect_manifest_urls(manifest):
        work_file = pathlib.Path(work_file)
        assert work_file.exists(), f"File {work_file} does not exist"
        yield work_file


def _vcf_item_generator(manifest: Manifest) -> Generator[tuple, None, None]:
    """Return a VCFItem for each line in the vcf."""
    total_lines = 0
    for work_file in tqdm(
        _work_file_generator(manifest),
        total=len(manifest.vcf_files),
        disable=manifest.disable_progress_bars,
    ):
        line_number = 0
        if "gz" in str(work_file):
            f = gzip.open(work_file, "rt")
        else:
            f = open(work_file, "r")
        with f:
            key = str(work_file)
            metrics[key][STATUS] = "started"
            metrics[key][START_TIME] = time.time()
            metrics[key][SUCCESSES] = 0
            metrics[key][METAKB_HITS] = 0

            for line in f:
                if line.startswith("#"):
                    continue

                line_number += 1
                total_lines += 1

                for gnomad_id in generate_gnomad_ids(
                    line, compute_for_ref=manifest.compute_for_ref
                ):
                    yield VCFItem(
                        fmt="gnomad",
                        var=gnomad_id,
                        file_name=work_file,
                        line_number=line_number,
                    )  # {"fmt": "gnomad", "var": gnomad_id}, work_file, line_number

                if manifest.limit and line_number > manifest.limit:
                    _logger.info(f"Limit of {manifest.limit} reached, stopping")
                    break

            _logger.info(f"Setting metrics for {work_file}")
            metrics[key][STATUS] = "finished"
            metrics[key][END_TIME] = time.time()
            metrics[key][LINE_COUNT] = line_number
            metrics[key][ELAPSED_TIME] = (
                metrics[key][END_TIME] - metrics[key][START_TIME]
            )

    _logger.info(
        f"_vcf_generator: Finished processing all files in the manifest {total_lines} lines processed."
    )


def _vrs_generator(manifest: Manifest) -> Generator[dict, None, None]:
    """Return a generator for the VRS ids."""
    tlr = Translator(normalize=manifest.normalize)
    c = 0
    for result in tlr.translate_from(
        generator=tqdm(
            _vcf_item_generator(manifest),
            total=manifest.estimated_vcf_lines,
            disable=manifest.disable_progress_bars,
        ),
        num_threads=manifest.num_threads,
    ):
        yield result
        c += 1
    _logger.info(
        f"_vrs_generator: Finished processing all vrs results in the manifest {c} results processed."
    )


def vrs_ids(allele: VRS.Allele) -> list[str]:
    """Return a list of VRS ids from an allele."""
    return [allele.id]  # , allele.location.id, allele.location.sequence_id]


def annotate_all(
    manifest: Manifest, max_errors: int, timestamp_str: str = None
) -> pathlib.Path:
    """Annotate all the files in the manifest. Return a file with metrics."""

    # set the manifest in a well known place, TODO: is this really necessary
    _logger.info("annotate_all: Starting.")
    vrs_anvil.manifest = manifest
    metakb_proxy = vrs_anvil.MetaKBProxy(
        metakb_path=pathlib.Path(manifest.metakb_directory),
        cache_path=pathlib.Path(manifest.cache_directory),
    )
    _logger.info("annotate_all: completed metakb init.")

    metrics[TOTAL][START_TIME] = time.time()
    total_errors = 0
    for result in _vrs_generator(manifest):
        assert result is not None, "result is None"
        assert isinstance(result, VCFItem), "result is not a VCFItem"

        file_path = str(result.file_name)

        if ERROR in result:
            errors = metrics[file_path][ERROR]
            if result[ERROR] not in errors:
                errors[result[ERROR]] = 0
            errors[result[ERROR]] += 1
            total_errors += 1
            if total_errors > max_errors:
                break
        else:
            allele_id = result.result

            metrics[file_path][SUCCESSES] += 1

            # check metaKB cache, TODO - it would be nice if we had the metakb.study.id and added that to result_dict
            if metakb_proxy.get(allele_id):
                _logger.info(f"VRS id {allele_id} found in metakb. {result}")

                # add vrs_id, allele_dict, actual evidence to this object as well (#3)
                metrics[file_path][MATCHES][allele_id] = {
                    "fmt": result.fmt,
                    "var": result.var,
                }

                metrics[file_path][METAKB_HITS] += 1

    _logger.info("annotate_all: Finished processing results.")

    metrics[TOTAL][TIMESTAMP] = timestamp_str
    metrics[TOTAL][END_TIME] = time.time()
    metrics[TOTAL][ELAPSED_TIME] = metrics[TOTAL][END_TIME] - metrics[TOTAL][START_TIME]
    metrics[TOTAL][SUCCESSES] = sum(
        [metrics[key].get(SUCCESSES, 0) for key in metrics.keys() if key != TOTAL]
    )
    metrics[TOTAL]["errors"] = sum(
        [sum(metrics[key]["errors"].values()) for key in metrics.keys() if key != TOTAL]
    )

    _logger.info("annotate_all: Finished calculating metrics.")

    # Append timestamp or suffix to filename
    if not timestamp_str:
        timestamp_str = datetime.now().strftime("%Y%m%d_%H%M%S")
    metrics_file = (
        pathlib.Path(manifest.state_directory) / f"metrics_{timestamp_str}.yaml"
    )
    with open(metrics_file, "w") as f:
        # clean up the recursive dict into a plain old dict so that it serialized to yaml neatly
        for k, v in metrics.items():
            metrics[k] = dict(v)
            if k != TOTAL:
                if "errors" in metrics[k]:
                    metrics[k]["errors"] = dict(metrics[k]["errors"])
                if MATCHES in metrics[k]:
                    metrics[k][MATCHES] = dict(metrics[k][MATCHES])

        yaml.dump(dict(metrics), f)

    _logger.info("annotate_all: Finished writing metrics.")

    return metrics_file



import os
from pathlib import Path
import sqlite3

from datetime import datetime
from ga4gh.va_spec.base.core import CohortAlleleFrequencyStudyResult as CAF
from ga4gh.va_spec.base.core import DataSet, StudyGroup
from pysam import VariantFile, VariantRecord
from plugin_system.plugins.base_plugin import BasePlugin

# location to register plugins classes
PLUGIN_MODULE_PATH = "plugin_system.plugins"


def get_cohort_allele_frequency(
    variant_id: str,
    vcf_path: str,
    vcf_index_path: str | None = None,
    participant_list: list[str] | None = None,
    phenotype: str | None = None,
    plugin: BasePlugin | None = None,
) -> CAF:
    """Create a cohort allele frequency for either genotypes or phenotypes

    Args:
        variant_id (str): variant ID (VRS ID)
        vcf_path (str): path to VCF
        vcf_index_path (str): path to VRS to VCF coordinates index (SQLite table)
        phenotype_table (str, optional): where to pull phenotype information from. Defaults to None.
        participant_list (list[str], optional): Subset of participants to use. Defaults to None.
        phenotype (str, optional): Specific phenotype to subset on. Defaults to None.
        plugin (BasePlugin, optional): Plugin object to use for custom processing. Defaults to None, loading in the BasePlugin.

    Returns:
        dict: Cohort Allele Frequency object
    """

    # check variant_id is VRS ID
    assert (
        "ga4gh:VA" in variant_id
    ), "variant ID type not yet supported, use VRS ID instead"

    # use default plugin if none specified
    if plugin is None:
        plugin = BasePlugin()

    # get index of variant to patient
    # in this case, the VCF row of the variant_id
    vcf = VariantFile(vcf_path)
    record = get_vcf_row(variant_id, vcf, vcf_index_path)

    # if multiple alts, get index associated with alt allele
    # if ref has been saved in VRS Alleles IDs, adjust indices to match
    alt_index = record.info["VRS_Allele_IDs"].index(variant_id)
    if "REF" not in vcf.header.info["VRS_Allele_IDs"].description:
        alt_index -= 1

    # create cohort, defaults to all samples listed in VCF
    cohort = (
        set(participant_list) if participant_list is not None else set(record.samples)
    )

    # variables for cohort allele frequency (CAF) object
    focus_allele_count = 0
    locus_allele_count = 0
    cohort_phenotypes = set() if phenotype is None else [phenotype]

    # only relevat if the variant is diploid
    num_homozygotes = 0
    num_hemizygotes = 0

    # aggregate data for CAF so long as...
    for sample_id, genotype in record.samples.items():
        # 1. sample genotype call exist
        alleles = genotype.allele_indices
        if all(a is None for a in alleles):
            continue

        # 2. sample in specified cohort
        if sample_id not in cohort:
            continue

        # 3. matches subcohort criteria
        should_include_sample = plugin.include_sample(sample_id, record, phenotype)
        if phenotype is not None and not should_include_sample:
            continue

        # with these conditions satisfied...
        num_focus_alleles, num_locus_alleles = plugin.process_sample_genotype(
            sample_id, record, alt_index
        )

        # increment allele counts
        focus_allele_count += num_focus_alleles
        locus_allele_count += num_locus_alleles

        # record zygosity
        if num_focus_alleles == 1:
            num_hemizygotes += 1
        elif num_focus_alleles == 2:
            num_homozygotes += 1

        # update phenotypes as necessary
        if phenotype is not None:
            continue
        else:
            # aggregate phenotypes if they exist
            phenotype_index = plugin.get_phenotype_index()
            if phenotype_index is not None and sample_id in phenotype_index:
                cohort_phenotypes.update(phenotype_index[sample_id])

    # format caf fields before populating caf object
    allele_frequency = (
        focus_allele_count * 1.0 / locus_allele_count if locus_allele_count != 0 else 0
    )

    if phenotype is None:
        cohort = StudyGroup(id="ALL", name="Overall")
    else:
        cohort = StudyGroup(id=phenotype, name=phenotype)

    ancillary_results = {
        "homozygotes": num_homozygotes,
        "hemizygotes": num_hemizygotes,
        "phenotypes": list(cohort_phenotypes),
    }

    # populate final caf object according to va-spec-python
    caf = CAF(
        sourceDataSet=DataSet(id=vcf_path, description=f"Created {datetime.now()}"),
        focusAllele=variant_id,
        focusAlleleCount=focus_allele_count,
        focusAlleleFrequency=allele_frequency,
        locusAlleleCount=locus_allele_count,
        cohort=cohort,
        ancillaryResults=ancillary_results,
    )

    return caf


def fetch_by_vrs_ids(
    vrs_ids: list[str], db_location: Path | None = None
) -> list[tuple]:
    """Access index by VRS ID.

    :param vrs_id: VRS allele hash (i.e. everything after ``"ga4gh:VA."``)
    :param db_location: path to sqlite file (assumed to exist)
    :return: location description tuple if available
    """

    trunc_vrs_ids = []
    for vrs_id in vrs_ids:
        trunc_vrs_id = vrs_id[9:] if vrs_id.startswith("ga4gh:VA.") else vrs_id
        trunc_vrs_ids.append(trunc_vrs_id)

    if not db_location.exists():
        raise OSError(f"Index at {db_location} does not exist")

    conn = sqlite3.connect(db_location)

    # have to manually make placeholders for python sqlite API --
    # should still be safe against injection by using parameterized query
    placeholders = ",".join("?" for _ in trunc_vrs_ids)
    result = conn.cursor().execute(
        f"SELECT vrs_id, chr, pos FROM vrs_locations WHERE vrs_id IN ({placeholders})",  # noqa: S608
        trunc_vrs_ids,
    )
    data = result.fetchall()

    if len(data) == 0:
        raise Exception(
            f"No matching rows in the VCF index for the VRS IDs specified \n   - VRS IDs: {vrs_ids} \n   - Index path: {db_location}"
        )

    conn.close()
    return data


def get_vcf_row(
    variant_id: str, vcf: VariantFile, index_path: str = None
) -> VariantRecord:
    """given a variant id and annotated VCF, get the associated VCF row

    Args:
        variant_id (str): VRS ID for the variant of interest
        vcf (VariantFile): Pysam VariantFile
        index_path (str, optional): Index used to speed up search for variant. Defaults to iterating through VCF.

    Raises:
        Exception: outputs if no index is found

    Returns:
        VariantRecord: A Pysam VariantRecord (VCF row)
    """
    if "VRS_Allele_IDs" not in vcf.header.info:
        raise KeyError(
            "no VRS_Allele_IDs key in INFO found, "
            "please ensure that this is an VRS annotated VCF"
        )

    # try to populate from Bash env variable
    if not index_path:
        assert (
            "VRS_VCF_INDEX" in os.environ
        ), "no genotype index specified, no index path was provided nor was a variable name VRS_VCF_INDEX found."
        index_path = Path(os.environ.get("VRS_VCF_INDEX"))

    if index_path:
        # if index provided, use it to get VCF row
        index_path = Path(index_path)

        # find variant of interest
        for _, chr, pos in fetch_by_vrs_ids([variant_id], index_path):
            # TODO [ISSUE-103]: generalize VCF fixtures
            for record in vcf.fetch(chr, pos - 1, pos):
                if variant_id in record.info["VRS_Allele_IDs"]:
                    return record

        raise KeyError(f"no VCF row found matching variant ID {variant_id}")
    else:
        # otherwise, iterate through VCF
        for record in enumerate(vcf.fetch()):
            print(
                "no VCF index specified, iterating through VCF to locate variant of interest"
            )
            if variant_id in record.info["VRS_Allele_IDs"]:
                return record



import pysam

from abc import ABC


class BasePlugin(ABC):
    """
    Interface for caf generation plugins
    """

    __is_plugin__ = True

    def __init__(self, phenotype_index: dict[str, list[str]] | None = None):
        """Constructor to initialize a phenotype index mapping from sample id to the sample's phenotypes.
        Index example: {"sample_A": ["HP:0001263", "HP:0000002"], "sample_B": ["HP:0001263"]}

        Args:
            phenotype_index (dict[str, list[str]]): dictionary mapping from sample id to sample's phenotypes. Defaults to None, where only genotype queries can be made.
        """
        self.phenotype_index = phenotype_index

    def get_phenotype_index(self) -> dict[str, list[str]]:
        """getter for a dictionary mapping from each sample to the sample's phenotypes

        Returns:
            dict[str, list[str]]: index mapping from sample id to sample's phenotypes. For example: {"patient_A": ["lactose intolerance", "anxiety"], "patient_B": ["shortness of breath"]}
        """
        return self.phenotype_index

    def include_sample(
        self, sample_id: str, record: pysam.VariantRecord, phenotype: str | None
    ) -> bool:
        """determine whether to include a sample in the cohort allele frequency based on its variant data and phenotypic traits.


        Args:
            sample_id (str): sample_id used to uniquly identify a sample ID
            record (pysam.VariantRecord): PySam record object representing a VCF row
            phenotype (str): phenotype of interest, matching phenotype codes used in sample_phenotype_)index

        Raises:
            Exception: if user subsets by phenotype without a phenotype mapping (ie phenotype index)

        Returns:
            bool: whether to include the sample
        """

        # if no sample to phenotype mapping but phenotype specified, include all samples
        # else raise error that no phenotype mapping to make use of
        if self.phenotype_index is None:
            if phenotype is None:
                return True
            else:
                raise Exception(
                    "phenotype is specified but no phenotype mapping exists. Please instantiate the plugin with a phenotype_index"
                )

        # if phenotype mapping, return if sample has phenotype
        has_specified_phenotype = (
            sample_id in self.phenotype_index
            and phenotype in self.phenotype_index[sample_id]
        )

        return has_specified_phenotype

    def process_sample_genotype(
        self,
        sample_id: str,
        record: pysam.VariantRecord,
        alt_index: int,
    ) -> tuple[int, int]:
        """given a sample's genotype, return focus and locus allele counts

        Args:
            sample_id (str): sample_id used to uniquly identify a sample ID
            record (pysam.VariantRecord): pysam record object representing a VCF row
            sample_phenotype_index (dict[str, list[str]]): mapping from sample IDs to each sample list of phenotypes
            alt_index (int): index matching the variant of interest

        Returns:
            tuple[int, int]: number of focus (specified) alleles, then number of locus (total) alleles.
        """

        # increment focus allele count, handling multiple alts edge case
        # for example, if the alt of interest is at index 3, then a genotype of
        # (3,2) would have a 1 focus allele out of 2 total alleles
        alleles = record.samples[sample_id].allele_indices
        num_focus_alleles = sum(
            [1 for _, alt_number in enumerate(alleles) if alt_number == alt_index]
        )
        num_total_alleles = len(alleles)

        return num_focus_alleles, num_total_alleles



import os
from pathlib import Path
from pysam import VariantFile

from plugin_system.plugins.gregor_plugin import GregorPlugin
from plugin_system.utils import load_dict, save_dict


def test_gregor_plugin_creates_correct_phenotype_index(
    chrY_vcf_path: str, gregor_plugin: GregorPlugin
):
    """given all sample IDs in a VCF, phenotypes index should be created with
    the expected number of unique phenotypes"""

    vcf = VariantFile(chrY_vcf_path)
    assert "VRS_Allele_IDs" in vcf.header.info, (
        "no VRS_Allele_IDs key in INFO found, "
        "please ensure that this is an VRS annotated VCF"
    )

    for record in vcf.fetch():
        phenotypes_set = set()
        samples = [sample for sample in record.samples]
        phenotype_index = gregor_plugin.get_phenotype_index()

        for sample_id in samples:
            if sample_id in phenotype_index:
                phenotypes_set.update(phenotype_index[sample_id])
        print("len(phenotypes_set):", len(phenotypes_set))

        expected_num_phenotypes = 980
        assert (
            len(phenotypes_set) == expected_num_phenotypes
        ), f"Expected {expected_num_phenotypes} phenotypes, got {len(phenotypes_set)}"

        break


def test_loading_gregor_phenotype_index_by_path(
    gregor_plugin: GregorPlugin, tmp_path: Path
):
    index = gregor_plugin.get_phenotype_index()
    save_path = tmp_path / "index.json"
    save_dict(index, save_path)

    loaded_index = load_dict(save_path)

    assert (
        index == loaded_index
    ), "saved index does not match loaded index... use -vv flag for a better diff"
    os.remove(save_path)



# plugin_system/plugin_manager.py
import pkgutil
import importlib
import os

from plugin_system.plugins.base_plugin import BasePlugin
from vrs_anvil.evidence import PLUGIN_MODULE_PATH


class PluginManager:
    def load_plugin(self, plugin_name: str) -> BasePlugin:
        """grab first plugin class matching name `plugin_name` within the `plugin_package` directory

        Args:
            plugin_name (str): name of the plugin class

        Raises:
            OSError: unable to find class named `plugin_name`
            ImportError: unable to import package

        Returns:
            BasePlugin: non-instantiated Plugin class
        """

        # get full path for the default plugin directory
        default_plugin_dir = os.path.dirname(
            importlib.import_module(PLUGIN_MODULE_PATH).__file__
        )

        # look for plugin by name first in default directory then in top-level directory
        for i, iter in enumerate(
            [pkgutil.iter_modules([default_plugin_dir]), pkgutil.iter_modules()]
        ):
            for _, name, _ in iter:
                # only look for specific file names
                if not name.endswith("_plugin"):
                    continue

                # get full module name, use only plugin name if iterating across top-level dirs
                module_name = f"{PLUGIN_MODULE_PATH}.{name}" if i == 0 else name

                try:
                    # dynamically import the module
                    module = importlib.import_module(module_name)

                    # grab first plugin class matching name <plugin_name>
                    for attribute_name in dir(module):
                        if attribute_name == plugin_name:
                            attribute = getattr(module, attribute_name)
                            if isinstance(attribute, type) and hasattr(
                                attribute, "__is_plugin__"
                            ):
                                return attribute
                except ImportError as e:
                    # plugin modules found but unable to load specified plugin
                    print(f"Error loading plugin {name}: {e}")
                    raise

        # if no plugin found, raise error
        raise OSError(
            f"Plugin {plugin_name} not found. Make sure the path is stored in the top-level"
        )



import os
import pathlib
from typing import Generator
import requests
from concurrent.futures import ThreadPoolExecutor, as_completed
from vrs_anvil import Manifest
from google.cloud import storage
import boto3


# TODO - not tested
def download_s3_object(bucket_name, object_name, destination_file_name) -> str:
    """Download an object from an S3 bucket and save it to a file."""
    if os.path.exists(destination_file_name):
        return destination_file_name
    s3 = boto3.client("s3")
    s3.download_http_file(bucket_name, object_name, destination_file_name)
    return destination_file_name


def download_google_blob(bucket_name, source_blob_name, destination_file_name) -> str:
    """Downloads a blob from the bucket."""
    if os.path.exists(destination_file_name):
        return destination_file_name

    storage_client = storage.Client()
    assert "GOOGLE_PROJECT" in os.environ, "GOOGLE_PROJECT environment variable not set"
    bucket = storage_client.bucket(
        bucket_name, user_project=os.getenv("GOOGLE_PROJECT")
    )
    blob = bucket.blob(source_blob_name)
    pathlib.Path(destination_file_name).parent.mkdir(parents=True, exist_ok=True)
    blob.download_to_filename(destination_file_name)
    return destination_file_name


# TODO - not tested
def download_http_file(url, destination_dir) -> str:
    """Download a file from a URL and save it to a directory."""
    filename = os.path.join(destination_dir, url.split("/")[-1])
    # check if file already exists
    if os.path.exists(filename):
        return filename
    response = requests.get(url)
    with open(filename, "wb") as f:
        f.write(response.content)
    return filename


def create_symlink_to_work_directory(work_directory: str, vcf_file: str) -> str:
    """Create a link to the work directory for vcf file."""
    # Extract the filename from the vcf_file path

    vcf_filename = os.path.basename(vcf_file)

    # Construct the full path for the symlink
    symlink_path = os.path.join(work_directory, vcf_filename)

    # Delete if exists (prevents cached invalid links)
    if os.path.islink(symlink_path):
        os.unlink(symlink_path)

    # Create the symlink
    vcf_file = os.path.abspath(vcf_file)
    os.symlink(vcf_file, symlink_path)

    return symlink_path


def collect_manifest_urls(manifest: Manifest) -> Generator[str, None, None]:
    """Collect the URLs from the manifest and download them."""
    # TODO - is this really too many threads? each download is IO bound
    with ThreadPoolExecutor(max_workers=max(len(manifest.vcf_files), 8)) as executor:
        futures = []
        for vcf_file in manifest.vcf_files:
            if vcf_file.startswith("http"):
                futures.append(
                    executor.submit(
                        download_http_file, vcf_file, manifest.work_directory
                    )
                )
            elif vcf_file.startswith("gs://"):
                bucket_name, blob_name = vcf_file[5:].split("/", 1)
                destination_file_name = os.path.join(manifest.work_directory, blob_name)
                futures.append(
                    executor.submit(
                        download_google_blob,
                        bucket_name,
                        blob_name,
                        destination_file_name,
                    )
                )
            elif vcf_file.startswith("s3://"):
                bucket_name, object_name = vcf_file[5:].split("/", 1)
                destination_file_name = os.path.join(
                    manifest.work_directory, object_name
                )
                futures.append(
                    executor.submit(
                        download_s3_object,
                        bucket_name,
                        object_name,
                        destination_file_name,
                    )
                )
            elif vcf_file.startswith("file://"):
                futures.append(
                    executor.submit(
                        create_symlink_to_work_directory,
                        manifest.work_directory,
                        vcf_file[7:],
                    )
                )
            else:
                futures.append(
                    executor.submit(
                        create_symlink_to_work_directory,
                        manifest.work_directory,
                        vcf_file,
                    )
                )
        # Yield results as they become available
        for _ in as_completed(futures):
            yield _.result()



import pathlib
import time
from typing import Generator

import pytest

from tests.unit import validate_threaded_result
from vrs_anvil.translator import VCFItem

# see https://github.com/ga4gh/vrs-python/blob/main/tests/extras/test_allele_translator.py#L17

snv_inputs = {
    "hgvs": "NC_000019.10:g.44908822C>T",
    "beacon": "19 : 44908822 C > T",
    "spdi": "NC_000019.10:44908821:1:T",
    "gnomad": "19-44908822-C-T",
}

snv_output = {
    "id": "ga4gh:VA.0AePZIWZUNsUlQTamyLrjm2HWUw2opLt",
    "type": "Allele",
    "digest": "0AePZIWZUNsUlQTamyLrjm2HWUw2opLt",
    "location": {
        "id": "ga4gh:SL.wIlaGykfwHIpPY2Fcxtbx4TINbbODFVz",
        "type": "SequenceLocation",
        "digest": "wIlaGykfwHIpPY2Fcxtbx4TINbbODFVz",
        "sequenceReference": {
            "type": "SequenceReference",
            "refgetAccession": "SQ.IIB53T8CNeJJdUqzn9V_JnRtQadwWCbl",
        },
        "start": 44908821,
        "end": 44908822,
    },
    "state": {"type": "LiteralSequenceExpression", "sequence": "T"},
}

# https://www.ncbi.nlm.nih.gov/clinvar/variation/1373966/?new_evidence=true
deletion_inputs = {
    "hgvs": "NC_000013.11:g.20003097del",
    "spdi": ["NC_000013.11:20003096:C:", "NC_000013.11:20003096:1:"],
    "gnomad": "13-20003096-AC-A",
}

gnomad_deletion_output = {
    "id": "ga4gh:VA.agNTkBOVZ0HQP5iWJ2GoRovSaHUCyZHN",
    "type": "Allele",
    "digest": "agNTkBOVZ0HQP5iWJ2GoRovSaHUCyZHN",
    "location": {
        "id": "ga4gh:SL.30MgiQJ3IB1mk1wq4RyhUADF_V0vw0fD",
        "type": "SequenceLocation",
        "digest": "30MgiQJ3IB1mk1wq4RyhUADF_V0vw0fD",
        "sequenceReference": {
            "type": "SequenceReference",
            "refgetAccession": "SQ._0wi-qoDrvram155UmcSC-zA5ZK4fpLT",
        },
        "start": 20003096,
        "end": 20003097,
    },
    "state": {
        "type": "ReferenceLengthExpression",
        "length": 0,
        "sequence": "",
        "repeatSubunitLength": 1,
    },
}

# https://www.ncbi.nlm.nih.gov/clinvar/variation/1687427/?new_evidence=true
insertion_inputs = {
    "hgvs": "NC_000013.11:g.20003010_20003011insG",
    "spdi": ["NC_000013.11:20003010::G", "NC_000013.11:20003010:0:G"],
    "gnomad": "13-20003010-A-AG",
}

gnomad_insertion_output = {
    "digest": "YSFR_-q58UM-bsgyq9ScHa4hfNOWGelM",
    "id": "ga4gh:VA.YSFR_-q58UM-bsgyq9ScHa4hfNOWGelM",
    "location": {
        "digest": "XHzVIQ1JEps1jDJIMGF_kn7YIcQdn5bW",
        "id": "ga4gh:SL.XHzVIQ1JEps1jDJIMGF_kn7YIcQdn5bW",
        "end": 20003010,
        "start": 20003010,
        "sequenceReference": {
            "refgetAccession": "SQ._0wi-qoDrvram155UmcSC-zA5ZK4fpLT",
            "type": "SequenceReference",
        },
        "type": "SequenceLocation",
    },
    "state": {"sequence": "G", "type": "LiteralSequenceExpression"},
    "type": "Allele",
}

# https://www.ncbi.nlm.nih.gov/clinvar/variation/1264314/?new_evidence=true
duplication_inputs = {
    "hgvs": "NC_000013.11:g.19993838_19993839dup",
    "spdi": "NC_000013.11:19993837:GT:GTGT",
    "gnomad": "13-19993838-GT-GTGT",
}

duplication_output = {
    "digest": "2Ju2sXpBNposOefRvorsmfqkAvt9tRHD",
    "id": "ga4gh:VA.2Ju2sXpBNposOefRvorsmfqkAvt9tRHD",
    "location": {
        "digest": "yCVGYQzbSLQe-GeAaHbW0dOiEGzHF3Yj",
        "id": "ga4gh:SL.yCVGYQzbSLQe-GeAaHbW0dOiEGzHF3Yj",
        "end": 19993839,
        "start": 19993837,
        "sequenceReference": {
            "refgetAccession": "SQ._0wi-qoDrvram155UmcSC-zA5ZK4fpLT",
            "type": "SequenceReference",
        },
        "type": "SequenceLocation",
    },
    "state": {
        "length": 4,
        "repeatSubunitLength": 2,
        "sequence": "GTGT",
        "type": "ReferenceLengthExpression",
    },
    "type": "Allele",
}


def test_normalized_results(caching_translator):
    tlr = caching_translator
    tlr.normalize = True

    assert (
        tlr._from_gnomad(snv_inputs["gnomad"]).model_dump(exclude_none=True)
        == snv_output
    )
    assert (
        tlr._from_gnomad(deletion_inputs["gnomad"]).model_dump(exclude_none=True)
        == gnomad_deletion_output
    )
    assert (
        tlr._from_gnomad(insertion_inputs["gnomad"]).model_dump(exclude_none=True)
        == gnomad_insertion_output
    )
    assert (
        tlr._from_gnomad(duplication_inputs["gnomad"]).model_dump(exclude_none=True)
        == duplication_output
    )


def test_results(caching_translator):
    """Ensure we can get the same results from gnomad as vrs-python, ie
    that the id and digest are computed recursively for the Allele object"""
    tlr = caching_translator
    assert tlr is not None

    inputs_dict = {
        "snv": (snv_inputs, snv_output),
        "deletion": (deletion_inputs, gnomad_deletion_output),
        "insertion": (insertion_inputs, gnomad_insertion_output),
        "duplication": (duplication_inputs, duplication_output),
    }

    for variant_type, (input, expected_allele) in inputs_dict.items():
        gnomad_expr = input["gnomad"]
        # allele object validation
        allele_id = tlr.translate_from(fmt="gnomad", var=gnomad_expr)

        assert expected_allele["id"] == allele_id, (
            f"{allele_id} does not match for {gnomad_expr}: "
            + f"\nexpected: {expected_allele['id']}"
        )


def test_cache(caching_translator):
    """Ensure that results from getting allele IDs are faster the second time."""
    tlr = caching_translator
    assert tlr is not None
    all_inputs = [snv_inputs, deletion_inputs, insertion_inputs, duplication_inputs]

    start_time = time.time()
    for inputs in all_inputs:
        tlr.translate_from(fmt="gnomad", var=inputs["gnomad"])
    noncached_time = time.time() - start_time

    start_time = time.time()
    for inputs in all_inputs:
        tlr.translate_from(fmt="gnomad", var=inputs["gnomad"])
    cache_time = time.time() - start_time

    assert cache_time < (
        noncached_time / 4
    ), f"Cache should make things significantly faster first {noncached_time} second {cache_time}."


@pytest.fixture()
def num_threads():
    """Return the number of threads to use for testing."""
    # TODO - anything more than 2 produces errors, need to investigate
    # e.g.  {'error': 'bad parameter or other API misuse', 'item': {'fmt': 'gnomad', 'var': '13-20003010-A-AG'}}
    return 20


def test_threading(translator, num_threads):
    """Ensure we can feed the threaded_translate_from method a generator and get results back."""
    tlr = translator
    assert tlr is not None

    parameters = [
        {"fmt": "gnomad", "var": snv_inputs["gnomad"]},
        {"fmt": "gnomad", "var": deletion_inputs["gnomad"]},
        {"fmt": "gnomad", "var": insertion_inputs["gnomad"]},
        {"fmt": "gnomad", "var": duplication_inputs["gnomad"]},
    ]

    def repeat_sequence(sequence, times):
        """
        Generator that repeats a given sequence for a specified number of times.

        Args:
        - sequence: The sequence to be repeated.
        - times: The number of times to repeat the sequence.

        Yields:
        - Elements from the sequence in a repeated pattern.
        """
        line_number = 1
        for _ in range(times):
            for _sequence in sequence:
                yield VCFItem(
                    **_sequence,
                    file_name="test",
                    line_number=line_number,
                    identifier=None,
                )
                line_number += 1

    c = 0  # count of results
    _times = 2
    for result in tlr.translate_from(
        repeat_sequence(parameters, times=_times), num_threads=num_threads
    ):
        c += 1
        validate_threaded_result(result, validate_passthrough=False)


@pytest.fixture
def gnomad_csv() -> pathlib.Path:
    """Return a path to a gnomad csv file."""
    _ = pathlib.Path(
        "tests/fixtures/gnomAD_v4.0.0_ENSG00000012048_2024_03_04_18_33_26.csv"
    )
    assert _.exists()
    return _


def gnomad_ids(path, limit=None) -> Generator[tuple, None, None]:
    """Open the csv file and yield the first column 'gnomAD ID', skip the header"""
    c = 0
    with open(path, "r") as f:
        skip = True
        for line in f:
            if skip:
                skip = False
                continue
            _ = line.split(",")
            gnomad_id = _[0]
            # use allele number as the identifier
            allele_number = _[18]
            yield VCFItem("gnomad", gnomad_id, path, c, allele_number)
            c += 1
            if limit and c > limit:
                break


def test_gnomad(translator, gnomad_csv, num_threads):
    """We can process a set of gnomad variants."""
    tlr = translator
    assert tlr is not None
    c = 0  # count of results
    for result_dict in tlr.translate_from(
        generator=gnomad_ids(gnomad_csv), num_threads=num_threads
    ):
        c += 1
        validate_threaded_result(result_dict, validate_passthrough=False)


def test_gnomad_inline(translator, gnomad_csv, num_threads=1):
    """We can process a set of gnomad variants."""
    tlr = translator
    assert tlr is not None
    c = 0  # count of results
    for result_dict in tlr.translate_from(
        generator=gnomad_ids(gnomad_csv), num_threads=num_threads
    ):
        c += 1
        validate_threaded_result(result_dict, validate_passthrough=False)



from plugin_system.plugins.base_plugin import BasePlugin


class SimplePlugin(BasePlugin):
    """
    Simple example plugin that automatically inherits all default methods from the BasePlugin.
    """



"""
Get statistics from 1000g samples, including
  - total samples
  - % matches to each kb individually and metakb as a whole
  - study ids and descriptions by variant hit in metakb

(to be run in a tmp/ directory)
"""

import io
import json
import os
import pandas as pd
import pysam
import subprocess
import yaml

from collections import defaultdict
from google.cloud import storage
from glob import glob
from firecloud import api as fapi
from vrs_anvil import query_metakb
from vrs_anvil.annotator import MATCHES, TOTAL, VRS_OBJECT

# data source
SOURCE_BUCKET = os.getenv("WORKSPACE_BUCKET").split("//")[1]
TIMESTAMP = "20240329_013953"
BUCKET_DIR = "scattered/state-04-07-2024"

# data directories
REPO_DIR = os.path.expanduser("~/vrs_anvil_toolkit")
BASE_DIR = f"{REPO_DIR}/tmp"
INPUT_DIR = f"{BASE_DIR}/work/high_coverage_3202_samples"
SEQREPO_DIR = "/home/jupyter/seqrepo/latest"
METRICS_DIR = os.path.expanduser(f"{BASE_DIR}/state/{TIMESTAMP}")

os.makedirs(INPUT_DIR, exist_ok=True)
os.makedirs(METRICS_DIR, exist_ok=True)

# ensure terra is setup
setup_cmd = f"cd {REPO_DIR} && bash terra/setup.sh"
output = subprocess.run(setup_cmd, shell=True, check=True)


# helper function
def truncate(s, first_few, last_few):
    "truncate string printing only first_few and last_few characters"
    return f"{s[:first_few]}...{s[-last_few:]}"


# get all metrics file blobs
client = storage.Client()
source_bucket = client.get_bucket(SOURCE_BUCKET)
all_blobs = source_bucket.list_blobs(prefix=BUCKET_DIR)
blobs = [blob for blob in all_blobs if blob.name.endswith(".yaml")]
print("number of metrics files:", len(blobs))

# download blobs to file if not already downloaded
print(f"downloading to {METRICS_DIR}:")
metrics_paths = []

for blob in blobs:
    file_name = blob.name.split("/")[-1]
    file_path = f"{METRICS_DIR}/{file_name}"
    metrics_paths.append(file_path)

    if os.path.exists(file_path):
        print(f"[SKIPPED] {file_name} already exists")
    else:
        blob.download_to_filename(file_path)

        print(f"downloaded {file_path}")

# get original 1000g metadata
df = pd.read_csv(
    io.StringIO(
        fapi.get_entities_tsv(
            "anvil-datastorage",
            "AnVIL_1000G_PRIMED-data-model",
            "sequencing_file",
            model="flexible",
        ).text
    ),
    sep="\t",
)

# get rid of gvcf data
df_vcf = df[df["file_type"].isin(["VCF", "VCF index"])]
df_1kgp = df_vcf[df_vcf["file_path"].str.contains("1kGP")]

num_vcf_idx_files = sum(df_1kgp["file_type"] == "VCF index")
num_vcf_files = sum(df_1kgp["file_type"] == "VCF")
assert (
    num_vcf_files == 23 and num_vcf_idx_files == 23
), f"check number of files, {num_vcf_files} vcfs and {num_vcf_idx_files} index files"

# load all vcf files if not pulled
chrs = [str(num) for num in range(1, 23)] + ["X"]

if chrs is not None:
    df_chrs = df_1kgp[df_1kgp["chromosome"].isin(set(chrs))]
else:
    df_chrs = df_1kgp
    specific_chromosomes = df_chrs["chromosome"].unique()

assert len(df_chrs) == 2 * len(
    chrs
), f"Expected 2 files per chr but {len(df_chrs)} files and {len(chrs)} chrs"

# pull from cloud and save paths and chr order
raw_vcfs = []
chrs_of_interest = []

for _, row in df_chrs.iterrows():
    uri = row["file_path"]
    file_name = uri.split("/")[-1]
    output_file = f"{INPUT_DIR}/{file_name}"

    # save vcfs for later, ignore indices
    if row["file_type"] == "VCF":
        raw_vcfs.append(output_file)
        chrs_of_interest.append(row["chromosome"])

    if os.path.exists(output_file):
        print(f"[SKIPPING] {truncate(file_name, 35, 10)} already exists")
    else:
        pull_vcf_cmd = f"gsutil -u $GOOGLE_PROJECT cp {uri} {INPUT_DIR}/"
        output = subprocess.run(pull_vcf_cmd, shell=True, check=True)

# collect all evidence from a list of metrics files in a directory
matches_per_file = {}
all_metrics_paths = sorted(glob(f"{METRICS_DIR}/metrics*.yaml"))

total_matches = 0
for metrics_path in all_metrics_paths:

    with open(metrics_path, "r") as file:
        metrics = yaml.safe_load(file)

    # for each file in the metrics, add matches to dict
    for file_path, metrics_dict in metrics.items():
        if file_path == TOTAL:
            continue

        # load original vcf
        original_path = os.path.expanduser(f"{BASE_DIR}/{file_path}")
        vcf_reader = pysam.VariantFile(original_path)
        num_samples = len(vcf_reader.header.samples)
        assert (
            num_samples == 3202
        ), f"expected 3202 samples for {file_path}, got {num_samples}"

        # check if matches to metakb cache found
        print("from file", truncate(original_path, 0, 46))
        if MATCHES in metrics_dict:
            expected_matches = int(metrics_dict["metakb_hits"])
            actual_matches = len(metrics_dict[MATCHES])
            assert (
                expected_matches == actual_matches
            ), f"expected {expected_matches} metakb matches, found {actual_matches}"
            print(f"{actual_matches} matches found")

            total_matches += actual_matches
            matches_per_file.update({original_path: metrics_dict[MATCHES]})
        else:
            print("no matches from this file")
        print()

print(f"total matches: {total_matches}")

# print all files with matches
print("files with evidence:")
for file_path in sorted(matches_per_file.keys()):
    print(truncate(file_path, 0, 47))


# function to update samples dict given coordinates and a vcf
def update_sample_evidence_dict(
    sample_evidence_dict, vcf_reader, chrom, pos, ref, alt, study_ids, variant_types
):
    """update sample evidence for a given allele"""

    found_corr_record = False

    if sum([1 for _ in vcf_reader.fetch(chrom, pos - 1, pos)]) != 1:
        print(f"\t[WARNING] more than one record found at pos {pos}")

    for _, record in enumerate(vcf_reader.fetch(chrom, pos - 1, pos)):
        print(
            f"\tparsing samples for row {record.chrom}, {record.pos}, {record.ref}, {record.alts}"
        )

        assert (
            not found_corr_record
        ), "more than 1 record corresponds to the same ref, alt"

        assert ref == record.ref, f"expected ref {record.ref}, got {ref}"
        if alt not in record.alts:
            print(
                f"\tSkipping {chrom}-{pos}-{ref}-{alt}, expected one of {record.alts} alts, got {alt}"
            )
            continue

        # populate patient/sample dictionary
        found_corr_record = True
        id_count = 0

        for sample, genotype in record.samples.items():
            gt_values = genotype["GT"]

            # Check if any allele is non-zero
            if any(int(allele) > 0 for allele in gt_values):
                id_count += 1
                sample_evidence_dict[sample] = {
                    "study_ids": list(study_ids),
                    "variant_types": list(variant_types),
                }

        return id_count


# get focus allele count for cohort allele frequency given a vcf and coordinates
def get_focus_allele_count(vcf_reader, chrom, pos, ref, alt):
    """calculates focus allele count for a given allele"""

    for _, record in enumerate(vcf_reader.fetch(chrom, pos - 1, pos)):
        print(
            f"\tCAF: getting allele counts for {record.chrom}, "
            + f"{record.pos}, {record.ref}, {record.alts}"
        )

        # check correct row
        assert ref == record.ref, f"expected ref {record.ref}, got {ref}"
        if alt not in record.alts:
            print(
                f"\tSkipping {chrom}-{pos}-{ref}-{alt}, expected one of "
                + f"{record.alts} alts, got {alt}"
            )
            continue

        # get number of alleles across all patients
        focus_allele_count = sum(
            [sum(genotype["GT"]) for _, genotype in record.samples.items()]
        )

    return focus_allele_count


# create a structured caf dict
def create_caf_dict(
    allele_id,
    gnomad_expression,
    focus_allele,
    focus_allele_count,
    locus_allele_count,
    ancillary_results=None,
):
    """create cohort allele frequency dict"""
    allele_frequency = focus_allele_count * 1.0 / locus_allele_count

    caf_dict = {
        "id": allele_id,
        "type": "CohortAlleleFrequency",
        "label": f"Overall Cohort Allele Frequency for {allele_id} ({gnomad_expression})",
        "derivedFrom": {
            "id": f"AnVIL_1000G_PRIMED-data-model",
            "type": "DataSet",
            "label": f"AnVIL 1000G PRIMED Data Model",
            "version": "Last Updated 3/12/2024",
        },
        "focusAllele": focus_allele,
        "focusAlleleCount": focus_allele_count,
        "locusAlleleCount": locus_allele_count,
        "alleleFrequency": allele_frequency,
        "cohort": {
            "id": "all_3202_samples",
            "label": "High-coverage sequencing of 3202 samples",
        },
    }

    if ancillary_results is not None:
        caf_dict["ancillaryResults"] = ancillary_results

    return caf_dict


# data stores
sample_evidence_dict = defaultdict(list)
variant_types_set = set()
caf_dicts = []
metakb_api_hits = []

# parse variant matches to save metakb study data
for file_path, matches in matches_per_file.items():
    vcf_reader = pysam.VariantFile(file_path)

    print(truncate(file_path, 0, 47))

    # for each variant per file
    for allele_id, allele_info in matches.items():
        sample_dict = defaultdict(list)

        # extract coordinate information
        gnomad_expr = allele_info["parameters"]["var"]
        chrom, pos, ref, alt = gnomad_expr.split("-")
        pos = int(pos)
        print(f"\tgnomad: {gnomad_expr}")
        print(f"\tallele_id: {allele_id}")

        # get study id associated with vrs allele id
        metakb_response = query_metakb(allele_id)
        if metakb_response is None:
            print(f"no metakb hit for allele {allele_id}\n")
        else:
            metakb_api_hits.append(allele_id)
            study_ids = metakb_response["study_ids"]

            assert study_ids == [
                study["id"] for study in metakb_response["studies"]
            ], "study ids in wrong order to add variant types"
            variant_types = [
                study["qualifiers"]["alleleOrigin"]
                for study in metakb_response["studies"]
            ]
            variant_types_set.update(variant_types)

            for study in metakb_response["studies"]:
                print(f"\t{study['type']} ({study['id']}): {study['description']}")
            print()

        id_count = update_sample_evidence_dict(
            sample_dict, vcf_reader, chrom, pos, ref, alt, study_ids, variant_types
        )

        print(f"\tNumber of matching samples: {id_count}\n")

        # get cohort allele counts
        focus_allele_count = get_focus_allele_count(vcf_reader, chrom, pos, ref, alt)
        print("\tFocus allele count: ", focus_allele_count)

        locus_allele_count = num_samples * 2
        allele_freq = focus_allele_count * 1.0 / (num_samples * 2)
        focus_allele = (
            allele_info[VRS_OBJECT] if VRS_OBJECT in allele_info else allele_id
        )

        # create ancillary_results for caf
        ancillary_results = {"patient_matches": id_count, "sample_dict": sample_dict}

        if metakb_response is not None:
            ancillary_results["metakb_dict"] = metakb_response

        # create and save caf dict
        caf_dict = create_caf_dict(
            allele_id,
            gnomad_expr,
            focus_allele,
            focus_allele_count,
            locus_allele_count,
            ancillary_results,
        )
        print(f"\tAdding {caf_dict['label']}\n")
        caf_dicts.append(caf_dict)

    print("~~~~~~~~~~~~~\n")

# output stats on cache to metakb api hits
metakb_cache_hits = sum(len(e) for e in matches_per_file.values())
print(
    f"ratio of metakb cache hits to api hits: {len(metakb_api_hits)}/{metakb_cache_hits}"
)
print(
    "Note: some metakb cache hits are not api hits because some VRS IDs may have matched molecular profiles",
    "that have no corresponding study ID in metakb because it did not pass metakb's normalizers",
)

# write Cohort Allele Frequency (CAF) objects to file, then read from it
caf_dir = os.path.expanduser(f"{BASE_DIR}/state")
os.makedirs(caf_dir, exist_ok=True)
file_name = f"caf_objects_{TIMESTAMP}.json"

with open(f"{caf_dir}/{file_name}", "w") as file:
    json.dump(caf_dicts, file)

print(f"\nwrote cohort allele frequency objects to {caf_dir}/{file_name}!")



import os
from pathlib import Path
import pytest

from pysam import VariantFile

from plugin_system.plugin_manager import PluginManager
from plugin_system.plugins.gregor_plugin import GregorPlugin
from plugin_system.utils import WORKSPACE_ENV_KEYS


@pytest.fixture
def chrY_vcf_path():
    """Return a GS URI to a remote VCF. Currently using annotated chrY GREGoR VCF"""

    return "gs://fc-secure-0b96edf7-d9b1-4ee2-94c9-c458fa44ccb1/gregor_joint_callset/gregor_consortium_u06_sorted_chrY_GRU_VRS.vcf.gz"


@pytest.fixture
def chr3_vcf_path():
    return "gs://fc-secure-0b96edf7-d9b1-4ee2-94c9-c458fa44ccb1/gregor_joint_callset/gregor_consortium_u06_sorted_chr3_GRU_VRS.vcf.gz"


@pytest.fixture()
def vrs_id_chr3(chr3_vcf_path):
    """VRS ID extracted from VCF row with only one alt"""

    chrom = "chr3"
    pos = 10456
    for record in VariantFile(chr3_vcf_path).fetch(chrom, pos - 1, pos):
        if record.ref == "CTT":
            return record.info["VRS_Allele_IDs"][1]  # 1 since 0 is a ref

    raise FileNotFoundError("couldn't find record for vrs_id_solo_alt")


@pytest.fixture
def vrs_vcf_index() -> str:
    tests_dir = Path(os.path.dirname(__file__)).parent.parent
    index_path = os.path.join(
        tests_dir.absolute(), "fixtures/gregor/chr3_chrY_index.db"
    )

    print("vrs_vcf_index path:", index_path)
    if os.path.exists(index_path):
        return str(index_path)

    assert (
        "PHENOTYPE_TABLE" in os.environ
    ), "No phenotype table found, see tests/fixtures/gregor/README.md for setup"

    return os.environ["VRS_VCF_INDEX"]


@pytest.fixture(scope="module")
def phenotype_table_path() -> str | None:
    if all([os.getenv(key) is not None for key in WORKSPACE_ENV_KEYS]):
        return None

    tests_dir = Path(os.path.dirname(__file__)).parent.parent
    path = os.path.join(tests_dir.absolute(), "fixtures/gregor/phenotype.tsv")

    print("phenotype_table_path:", path)
    if os.path.exists(path):
        return str(path)

    assert (
        "PHENOTYPE_TABLE" in os.environ
    ), "No phenotype table found, see tests/fixtures/gregor/README.md for setup"

    return os.environ["PHENOTYPE_TABLE"]


@pytest.fixture(scope="module")
def gregor_plugin(phenotype_table_path) -> GregorPlugin:
    # load plugin of choice
    plugin = PluginManager().load_plugin("GregorPlugin")

    # instantiate plugin
    return plugin(phenotype_table_path=phenotype_table_path)



from vrs_anvil import generate_gnomad_ids


def test_generate_gnomad_ids(caching_translator):
    input_vcf = "tests/fixtures/test_vcf_input.vcf"
    errors = []
    results = []
    with open(input_vcf, "r") as f:
        for line in f:
            if line.startswith("#"):
                continue
            for gnomad_id in generate_gnomad_ids(line):
                try:
                    result = caching_translator.translate_from(
                        fmt="gnomad", var=gnomad_id
                    )
                    results.append((gnomad_id, result))
                except Exception as e:
                    errors.append((gnomad_id, e))
    print(errors)
    assert len(results) >= 12, f"Errors: {len(errors)} Successes: {len(results)}"



import ast
import json
import logging
import multiprocessing
import os
import pickle
import subprocess
import pysam
import requests

from biocommons.seqrepo import SeqRepo
from datetime import datetime
from ga4gh.vrs import models
from ga4gh.vrs.dataproxy import SeqRepoDataProxy, create_dataproxy
from ga4gh.vrs.extras.translator import AlleleTranslator
from ga4gh.vrs.extras.annotator.vcf import VcfAnnotator
from pathlib import Path

# TODO: remove after adding MetaKB API query functionality


# get vrs ids
def translate(gnomad_expr):
    data_proxy = create_dataproxy()
    translator = AlleleTranslator(data_proxy)
    allele = translator._from_gnomad(gnomad_expr)
    return (gnomad_expr, dict(allele))


def annotate_vcf(path, require_validation=False):
    """param stem: path of input vcf file"""
    stem = path.replace(".vcf", "")

    input_vcf = path
    # output_vcf = f"{stem}.output.vcf.gz"
    output_vcf = ""
    output_pkl = f"{stem}-vrs-objects.pkl"

    vcf_annotator = VcfAnnotator(data_proxy=create_dataproxy())
    vcf_annotator.tlr.rle_seq_limit = None
    vcf_annotator.annotate(
        vcf_in=input_vcf,
        vcf_out=output_vcf,
        vrs_pickle_out=output_pkl,
        require_validation=require_validation,
    )

    return output_vcf, output_pkl


def print_dict(d, indent=2):
    """pretty print object as json"""
    print(json.dumps(d, indent=indent))


def print_percent(a, b):
    "pretty print percentages"
    print(f"{a}/{b} = {(100.0*a/b):.1f}%")


def annotate_vcf(
    input_vcf,
    output_vcf,
    output_pkl,
    seqrepo_root_dir,
    require_validation=True,
    rle_seq_limit=50,
):
    """param stem: path of input vcf file"""
    vcf_annotator = VCFAnnotator(seqrepo_root_dir=seqrepo_root_dir)
    vcf_annotator.tlr.rle_seq_limit = rle_seq_limit
    vcf_annotator.annotate(
        vcf_in=input_vcf,
        vcf_out=output_vcf,
        vrs_pickle_out=output_pkl,
        require_validation=require_validation,
    )

    return output_vcf, output_pkl


def unpickle(file_name):
    """Unpickle vrs objects to single dict"""
    with open(file_name, "rb") as f:
        vrs_objects = pickle.load(f)
        for k, v in vrs_objects.items():
            yield k, v


def get_num_variants(input_vcf):
    # get total num_variants
    return int(
        subprocess.run(
            f"grep -v '^#' {input_vcf} | wc -l",
            stdout=subprocess.PIPE,
            shell=True,
            check=True,
            text=True,
        ).stdout
    )


def parallelize(
    vrs_decorator, vrs_objects, worker_count, progress_interval=500, limit=None
):
    """harvest data from service"""

    manager = multiprocessing.Manager()
    results = manager.list()

    with multiprocessing.Pool(worker_count) as pool:
        # call the function for each item in parallel
        c = 0
        print(datetime.now().isoformat(), c)

        for result in pool.imap(vrs_decorator, vrs_objects):
            c += 1
            if result:
                results.append(result)
            if c == limit:
                break
            elif c % progress_interval == 0:
                print(datetime.now().isoformat(), c)

    return list(results)


def truncate(s, first_few, last_few):
    "truncate string printing only first_few and last_few characters"
    return f"{s[:first_few]}...{s[-last_few:]}"



import json
import os
import subprocess

from ga4gh.vrs.extras.translator import AlleleTranslator
from ga4gh.vrs.dataproxy import create_dataproxy
from plugin_system.plugin_manager import PluginManager
from vrs_anvil.evidence import get_cohort_allele_frequency


# run this in 1000g directory
assert os.getcwd().endswith(
    "1000g"
), "to ensure the plugin can be located, please run this in the 1000g directory"

# set varaible for variant data input
variant_id = "chr1-20094-TAA-T"
vcf_path = "../tests/fixtures/1kGP.chr1.1000.vrs.vcf.gz"

# set path to write VCF index to
vcf_index_path = "1000g_chr1_index.db"

# set phenotype-specific inputs
phenotype = "USA"  # to create subcohorts
phenotype_table = "population_descriptor.tsv"  # downloaded from https://anvil.terra.bio/#workspaces/anvil-datastorage/AnVIL_1000G_PRIMED-data-model/data

# create vcf index from vcf at the specified path using vrsix
command = ["vrsix", "load", f"--db-location={vcf_index_path}", vcf_path]
try:
    result = subprocess.run(command, check=True, text=True, capture_output=True)
    print("vrsix command executed successfully!")
except subprocess.CalledProcessError as e:
    print("Error executing vrsix command:", e.stderr)

# # get VRS ID from variant of interest
seqrepo_rest_service_url = "seqrepo+https://services.genomicmedlab.org/seqrepo"
seqrepo_dataproxy = create_dataproxy(uri=seqrepo_rest_service_url)
allele_translator = AlleleTranslator(seqrepo_dataproxy)
allele = allele_translator.translate_from(variant_id)
vrs_id = allele.id

# instantiate 1000G plugin class with phenotype table
plugin = PluginManager().load_plugin("ThousandGenomesPlugin")
simple_plugin = plugin(phenotype_table)

# generating cohort allele frequency using 1000G plugin
caf = get_cohort_allele_frequency(
    variant_id=vrs_id,
    vcf_path=vcf_path,
    vcf_index_path=vcf_index_path,
    plugin=simple_plugin,
    phenotype=phenotype,
)

print(f"CAF:")
print(json.dumps(caf.model_dump(exclude_none=True), indent=2))



def validate_threaded_result(result, validate_passthrough=False):
    """Test helper, a utility method to validate threaded lookup result."""
    assert result is not None, "result_dict is None"
    from vrs_anvil.translator import VCFItem

    assert isinstance(result, VCFItem), "result_dict is not a dict"

    allele_id = result.result
    assert isinstance(
        allele_id, str
    ), f"translated VRS Allele ID is a {type(allele_id)} not a string"
    prefix, hash = allele_id.split(".")
    assert (
        prefix == "ga4gh:VA" and len(hash) == 32
    ), "VRS Allele ID format has changed, consult https://vrs.ga4gh.org/en/stable/impl-guide/computed_identifiers.html#identify"

    for k in ["file_name", "line_number"]:
        if validate_passthrough:
            assert (
                getattr(result, k) is not None
            ), f"metrics tracking from caller {k} is None"



import io
import json
import os

from firecloud import api as fapi
import pandas as pd

####################################
#  utilities for transforming data #
####################################

WORKSPACE_ENV_KEYS = ["WORKSPACE_NAMESPACE", "WORKSPACE_NAME"]


def save_dict(d, path):
    # save to disk if specified
    if path:
        if os.path.exists(path):
            raise FileExistsError(
                f"index already exists at path: {path}. Please delete it before continuing"
            )
        else:
            with open(path, "w") as file:
                json.dump(d, file)


def load_dict(path):
    if os.path.exists(path):
        with open(path, "r") as file:
            d = json.load(file)
            return d
    else:
        raise FileNotFoundError(f"path to phenotype index does not exist: {path}")


def terra_data_table_to_dataframe(table_name) -> pd.DataFrame:
    # if unspecified, ensure valid Terra environment
    for env_key in WORKSPACE_ENV_KEYS:
        if env_key not in os.environ:
            raise OSError(
                f"ERROR: No {env_key} key found in environmental variables in the Terra workspace. If you are working in a Terra workspace, please ensure both a WORKSPACE_NAMESPACE and a WORKSPACE_NAME are specified."
            )

    # create dataframe from Terra data table
    # https://github.com/broadinstitute/fiss/blob/master/firecloud/api.py
    try:
        response = fapi.get_entities_tsv(
            os.environ["WORKSPACE_NAMESPACE"],
            os.environ["WORKSPACE_NAME"],
            table_name,
            model="flexible",
        )
        response.raise_for_status()
    except Exception as e:
        if response.json() and e in response.json():
            error_message = response.json()["message"]
        else:
            error_message = e
        print(
            f"Error while loading phenotype data table from workspace: \n{error_message}"
        )

    phenotype_tsv = io.StringIO(response.text)
    return pd.read_csv(phenotype_tsv, sep="\t")


def csv_to_dataframe(path):
    # table path specified, parse using that table
    with open(path, "r") as file:
        if path.endswith(".csv"):
            return pd.read_csv(file)
        elif path.endswith(".tsv"):
            return pd.read_csv(file, sep="\t")
        else:
            raise Exception(
                "Only csv and tsv file types implemented for phenotype table"
            )



import pytest

from vrs_anvil import METAKB_API, query_metakb
from metakb.schemas.api import SearchStatementsResponse


@pytest.fixture()
def vrs_id_1():
    return "ga4gh:VA.SOEVGpU16hxYQtJNeRyfq0V-B0rSOGK-"


def test_successful_query(vrs_id_1):
    response = query_metakb(vrs_id_1)
    ssr = SearchStatementsResponse(limit=None, **response)
    assert ssr is not None, "response is not a SearchStatementsResponse"
    assert response is not None, (
        f"unsuccessful query for {vrs_id_1}, "
        + f"ensure VRS ID digests have not changed at {METAKB_API}"
    )

    assert len(ssr.statement_ids) != 0, (
        "no study ids found, "
        + f"ensure response format or VRS ID digests have not changed at {METAKB_API}"
    )



from tests.unit import validate_threaded_result
from vrs_anvil import params_from_vcf


def test_1000g_vcf(translator, thousand_genome_vcf, num_threads):
    """Read and process AnVIL sourced vcf."""
    tlr = translator
    assert tlr is not None
    c = 0  # count of results
    for result_dict in tlr.translate_from(
        generator=params_from_vcf(thousand_genome_vcf), num_threads=num_threads
    ):
        c += 1
        print(result_dict)
        validate_threaded_result(result_dict, validate_passthrough=True)

    validate_threaded_result(result_dict)



import json
import os
import pytest

from typing import Generator
from pysam import VariantFile, VariantRecord
from pytest import approx

from vrs_anvil.evidence import get_cohort_allele_frequency

############
# FIXTURES #
############


@pytest.fixture
def chr3():
    """Return chromosome in the VCF."""
    return "chr3"


@pytest.fixture
def start():
    """Return the start range to query."""
    return 3188848


@pytest.fixture
def stop():
    """Return the end range to query."""
    return 3189029


@pytest.fixture
def expected_record_count():
    """Return the expected record count."""
    return 3


@pytest.fixture()
def vrs_id_chrY(chrY_vcf_path):
    """VRS ID extracted from VCF row with multiple alts"""

    for i, record in enumerate(VariantFile(chrY_vcf_path)):
        if i == 4:
            return record.info["VRS_Allele_IDs"][
                2
            ]  # index 2 refers to 2nd alt since 0 is a ref


def participants(record: VariantRecord) -> Generator[str, None, None]:
    """Return the participants that `have` this allele."""
    assert "GT" in record.format, "Genotype (GT) is required"

    for participant, values in record.samples.items():
        assert "GT" in values, "Genotype (GT) is required"
        # see https://samtools.github.io/hts-specs/VCFv4.1.pdf

        if any(values["GT"]):
            yield participant


#########
# TESTS #
#########


def test_remote_vcf(chrY_vcf_path, start, stop, expected_record_count):
    """Read a remote vcf file, query a range of alleles, check that at least 1 participant exists for each allele."""
    assert "GCS_OAUTH_TOKEN" in os.environ, (
        "GCS_OAUTH_TOKEN required: "
        "export GCS_OAUTH_TOKEN=$(gcloud auth application-default print-access-token)"
        "see https://github.com/pysam-developers/pysam/issues/592#issuecomment-353693674 https://support.terra.bio/hc/en-us/articles/360042647612-May-4-2020"
    )
    try:
        vcf = VariantFile(chrY_vcf_path)  # auto-detect input format
        # fetch returns pysam.libcbcf.VariantRecord
        records = [_ for _ in vcf.fetch("chrY", start, stop)]
        assert len(records) == expected_record_count

        for variant_record in records:
            my_participants = [_ for _ in participants(variant_record)]
            assert len(my_participants) < len(
                variant_record.samples
            ), "Not all participants have this allele"
            assert len(my_participants) > 0, "No participants have this allele"
    except ValueError as e:
        print("ValueError: has GCS_OAUTH_TOKEN expired?", e)
        raise e


def test_allele_counts_first_5_rows(chr3_vcf_path, vrs_vcf_index, gregor_plugin):
    """test that the calculated allele counts with no phenotype specified matches
    the actual counts stored in the INFO of the first 10 rows. Works for diploid (non-sex) variants
    """

    vcf = VariantFile(chr3_vcf_path)
    has_ref = "REF" in vcf.header.info["VRS_Allele_IDs"].description

    for i, record in enumerate(vcf.fetch()):
        print("~~~~~~~~~ row ", i, "~~~~~~~~~")

        # use only alt VRS IDs, not REFs
        vrs_allele_ids = record.info["VRS_Allele_IDs"]
        print("vrs_allele_ids:", vrs_allele_ids)
        if has_ref:
            vrs_allele_ids = vrs_allele_ids[1:]

        # for each alt ID, ensure stored allele counts match calculated allele counts
        for alt_index, allele_id in enumerate(vrs_allele_ids):
            print("alt id:", allele_id)
            if not allele_id:
                continue

            caf = get_cohort_allele_frequency(
                allele_id,
                chr3_vcf_path,
                vcf_index_path=vrs_vcf_index,
                plugin=gregor_plugin,
            )

            print("alt_index", alt_index)
            print("AC:", record.info["AC"][alt_index], caf.focusAlleleCount)
            print("AN:", record.info["AN"], caf.locusAlleleCount)
            focus_allele_count = record.info["AC"][alt_index]
            locus_allele_count = record.info["AN"]

            check_caf_allele_data(caf, focus_allele_count, locus_allele_count)
        if i == 5:
            break


def test_correct_caf_given_chr3_variant(
    vrs_id_chr3, chr3_vcf_path, vrs_vcf_index, gregor_plugin
):
    """test caf generation with default parameters and no phenotype specified"""

    # get and log caf
    caf = get_cohort_allele_frequency(
        vrs_id_chr3,
        chr3_vcf_path,
        vcf_index_path=vrs_vcf_index,
        plugin=gregor_plugin,
    )
    print_caf(caf)

    # sanity checks
    assert (
        caf.type == "CohortAlleleFrequencyStudyResult"
    ), f"object of type CohortAlleleFrequency not returned, returned {caf.type} instead"
    assert (
        caf.focusAlleleCount <= caf.locusAlleleCount
    ), f"Focus allele count ({caf['focusAlleleCount']}) is larger than locus allele count ({caf['locusAlleleCount']})"

    print("focusAlleleCount:", caf.focusAlleleCount)
    print("locusAlleleCount:", caf.locusAlleleCount)

    # check allele counts and frequency
    check_caf_allele_data(caf, expected_fac=183, expected_lac=896)

    # ensure list of phenotypes stored
    assert (
        "phenotypes" in caf.ancillaryResults
    ), "no phenotype key stored in caf.ancillaryResults"
    assert (
        len(caf.ancillaryResults["phenotypes"]) > 0
    ), 'no phenotypes stored in caf.ancillaryResults["phenotypes"]'


def test_correct_caf_given_chr3_variant_and_pheno(
    vrs_id_chr3, chr3_vcf_path, vrs_vcf_index, gregor_plugin
):
    """test caf generation for diploid variant with a specified phenotype"""

    phenotype = "HP:0001263"

    caf = get_cohort_allele_frequency(
        vrs_id_chr3,
        chr3_vcf_path,
        vcf_index_path=vrs_vcf_index,
        plugin=gregor_plugin,
        phenotype=phenotype,
    )
    print_caf(caf)

    # sanity checks
    assert (
        caf.type == "CohortAlleleFrequencyStudyResult"
    ), f"object of type CohortAlleleFrequency not returned, returned {caf.type} instead"
    assert (
        caf.focusAlleleCount <= caf.locusAlleleCount
    ), f"Focus allele count ({caf['focusAlleleCount']}) is larger than locus allele count ({caf['locusAlleleCount']})"

    print("focusAlleleCount:", caf.focusAlleleCount)
    print("locusAlleleCount:", caf.locusAlleleCount)

    # check focus counts, locus counts, and allele frequency
    check_caf_allele_data(caf=caf, expected_fac=2, expected_lac=26)

    # check phenotype is stored in cohort data
    assert phenotype in caf.cohort.id
    assert phenotype in caf.cohort.name


def test_correct_allele_freq_for_multi_alts_chrY_variant(
    vrs_id_chrY, chrY_vcf_path, vrs_vcf_index, gregor_plugin
):
    """for a vcf row with multiple alts, test caf generation with default parameters and no phenotype specified"""

    # creat caf
    caf = get_cohort_allele_frequency(
        vrs_id_chrY,
        chrY_vcf_path,
        vcf_index_path=vrs_vcf_index,
        plugin=gregor_plugin,
    )

    # logs
    print(f"CAF generated for {caf.focusAllele.root}")
    print("focusAlleleCount:", caf.focusAlleleCount)
    print("locusAlleleCount:", caf.locusAlleleCount)

    # check allele frequency
    expected_allele_freq = 0.0491
    actual_allele_freq = approx(caf.focusAlleleFrequency, abs=1e-4)
    assert (
        actual_allele_freq == expected_allele_freq
    ), f"incorrect allele frequency, expected {expected_allele_freq} got {actual_allele_freq}"

    # ensure list of phenotypes stored
    assert (
        "phenotypes" in caf.ancillaryResults
    ), "no phenotype key stored in caf.ancillaryResults"
    assert (
        len(caf.ancillaryResults["phenotypes"]) > 0
    ), 'no phenotypes stored in caf.ancillaryResults["phenotypes"]'


def test_correct_allele_freq_for_multi_alts_chrY_variant_and_phenotype(
    vrs_id_chrY, chrY_vcf_path, vrs_vcf_index, gregor_plugin
):
    """test caf generation specifying both a variant and a phenotype of interest"""

    phenotype = "HP:0001263"
    caf = get_cohort_allele_frequency(
        vrs_id_chrY,
        chrY_vcf_path,
        vcf_index_path=vrs_vcf_index,
        plugin=gregor_plugin,
        phenotype=phenotype,
    )
    print_caf(caf)

    expected_allele_freq = 0.1034
    actual_allele_freq = approx(caf.focusAlleleFrequency, abs=1e-4)
    assert (
        actual_allele_freq == expected_allele_freq
    ), f"incorrect allele frequency, expected {expected_allele_freq} got {actual_allele_freq}"


###########
# HELPERS #
###########


def check_caf_allele_data(caf, expected_fac, expected_lac):
    # get allele frequencies
    actual_af = approx(caf.focusAlleleFrequency, abs=1e-4)
    expected_af = expected_fac * 1.0 / expected_lac

    # set up values to compare
    actual_expected_pairs_with_name = [
        (caf.focusAlleleCount, expected_fac, "focusAlleleCount"),
        (caf.locusAlleleCount, expected_lac, "locusAlleleCount"),
        (actual_af, expected_af, "focusAlleleFrequency"),
    ]

    # assert calculated values are expected
    for actual, expected, name in actual_expected_pairs_with_name:
        assert (
            actual == expected
        ), f"incorrect {name}, expected {expected_fac} got {caf.focusAlleleFrequency}"


def print_caf(caf):
    print("CAF:")
    print(json.dumps(caf.model_dump(exclude_none=True), indent=2))



import logging
import pathlib
from typing import Generator

import pytest

from vrs_anvil.translator import threaded_translator, VCFItem

_logger = logging.getLogger("vrs_anvil.test_translator")


@pytest.fixture
def gnomad_csv() -> pathlib.Path:
    """Return a path to a gnomad csv file."""
    _ = pathlib.Path(
        "tests/fixtures/gnomAD_v4.0.0_ENSG00000012048_2024_03_04_18_33_26.csv"
    )
    assert _.exists()
    return _


def gnomad_ids(path, limit=None) -> Generator[VCFItem, None, None]:
    """Open the csv file and yield the first column 'gnomAD ID', skip the header"""
    c = 0
    with open(path, "r") as f:
        skip = True
        for line in f:
            if skip:
                skip = False
                continue
            _ = line.split(",")
            gnomad_id = _[0]
            # use allele number as the identifier
            allele_number = _[18]
            yield VCFItem("gnomad", gnomad_id, path, c, allele_number)
            c += 1
            if limit and c == limit:
                break


def test_threaded_translator(gnomad_csv):
    """Ensure threading works as expected."""

    limit = 2000
    num_worker_threads = 10

    generator_example = gnomad_ids(gnomad_csv, limit=limit)

    results = threaded_translator(generator_example, num_worker_threads)
    _logger.warning("test_threaded_translator: Starting test")

    # Add your assertions based on the expected results
    c = 0
    for _ in results:
        assert isinstance(_, VCFItem), "should get a VRS id"
        allele_id = _.result
        assert allele_id is not None, "allele.id is None"
        c += 1

    if limit:
        assert c == limit, "did not get the expected number of results"



import os
from datetime import datetime

import click
import yaml
import logging

from vrs_anvil import (
    Manifest,
    run_command_in_background,
    get_process_info,
    save_manifest,
)
from vrs_anvil.annotator import annotate_all
from logging.handlers import RotatingFileHandler
import pathlib

# Set up logging
log_format = "%(asctime)s %(threadName)s %(name)s [%(levelname)s] %(message)s"

_logger = logging.getLogger(__name__)


@click.group(invoke_without_command=True)
@click.version_option(package_name="vrs_anvil_toolkit")
@click.option(
    "--manifest",
    type=click.Path(exists=False),
    default="manifest.yaml",
    help="Path to manifest file.",
)
@click.option("--verbose", default=False, help="Log more information", is_flag=True)
@click.option("--max_errors", default=10, help="Number of acceptable errors.")
@click.option(
    "--suffix", default=None, help="Substitute timestamp with alternate file suffix"
)
@click.pass_context
def cli(ctx, verbose: bool, manifest: str, max_errors: int, suffix: str):
    """GA4GH GKS utility for AnVIL."""

    _log_level = logging.INFO
    if verbose:
        _log_level = logging.DEBUG

    if suffix is not None:
        timestamp_str = suffix
    else:
        timestamp_str = datetime.now().strftime("%Y%m%d_%H%M%S")

    try:
        with open(manifest, "r") as stream:
            manifest = Manifest.model_validate(yaml.safe_load(stream))

            # only create a persistent log for annotate subcommand
            if ctx.invoked_subcommand == "annotate":

                # Create a rotating file handler with a max size of 10MB and keep 3 backup files
                log_path = (
                    pathlib.Path(manifest.state_directory)
                    / f"vrs_anvil_{timestamp_str}.log"
                )
                file_handler = RotatingFileHandler(
                    log_path, maxBytes=10 * 1024 * 1024, backupCount=3
                )
                file_handler.setLevel(_log_level)
                file_handler.setFormatter(logging.Formatter(log_format))

                # Add the file handler to the logger
                # logger.addHandler(file_handler)
                # basicConfig call removed, which prevents the default configuration that logs to the console.
                logging.basicConfig(
                    level=_log_level, format=log_format, handlers=[file_handler]
                )

                click.secho(
                    f"🪵  Logging to {log_path}, level {logging.getLevelName(_log_level)}",
                    fg="yellow",
                )

            else:
                logging.basicConfig(level=_log_level, format=log_format)

            ctx.ensure_object(dict)
            ctx.obj["manifest"] = manifest
            ctx.obj["verbose"] = verbose
            ctx.obj["max_errors"] = max_errors
            ctx.obj["timestamp_str"] = timestamp_str

            if verbose:
                click.secho(f"📢  {manifest}", fg="green")
    except Exception as exc:
        click.secho(f"{exc}", fg="yellow")
        ctx.ensure_object(dict)


@cli.command("annotate")
@click.option(
    "--scatter",
    help="Start a background process per VCF file.",
    required=False,
    default=False,
    is_flag=True,
    show_default=True,
)
@click.pass_context
def annotate_cli(ctx, scatter: bool):
    """Read manifest file, annotate variants, all parameters controlled by manifest.yaml."""

    assert "manifest" in ctx.obj, "Manifest not found."
    timestamp_str = ctx.obj["timestamp_str"]

    # normal run with single process
    if not scatter:
        try:
            manifest = ctx.obj["manifest"]
            manifest_path = f"{manifest.work_directory}/manifest_{timestamp_str}.yaml"
            save_manifest(manifest, manifest_path)
            click.secho(f"🔑 Manifest saved at {manifest_path}", fg="yellow")
            _logger.debug(f"Manifest: {ctx.obj['manifest']}")

            click.secho("🚧  annotating variants", fg="yellow")
            metrics_file = annotate_all(
                manifest, max_errors=ctx.obj["max_errors"], timestamp_str=timestamp_str
            )
            click.secho(f"📊  metrics available in {metrics_file}", fg="green")
        except Exception as exc:
            click.secho(f"{exc}", fg="red")
            _logger.exception(exc)
    else:  # scattered processes / multiprocessing
        try:
            parent_manifest = ctx.obj["manifest"]
            scattered_processes = []
            child_processes = []

            for i, vcf_file in enumerate(parent_manifest.vcf_files):
                # create a new manifest for each VCF file based on the parent manifest
                child_manifest = parent_manifest.copy(deep=True)
                child_manifest.vcf_files = [vcf_file]
                child_manifest.num_threads = 1
                child_manifest.disable_progress_bars = True

                suffix_str = f"scattered_{timestamp_str}_{i}"
                child_manifest_path = (
                    pathlib.Path(child_manifest.work_directory)
                    / f"manifest_{suffix_str}.yaml"
                )
                save_manifest(child_manifest, child_manifest_path)
                click.secho(f"🔑 Manifest saved at {child_manifest_path}", fg="yellow")
                _logger.debug(f"Manifest: {ctx.obj['manifest']}")

                # run process to annotate each manifest
                process = run_command_in_background(
                    f"vrs_bulk --manifest {child_manifest_path} --suffix {suffix_str} annotate"
                )
                click.secho(
                    f"🚧  annotating {vcf_file} on pid {process.pid}", fg="yellow"
                )
                scattered_processes.append(
                    {
                        "pid": process.pid,
                        "manifest": str(child_manifest_path),
                        "vcf": vcf_file,
                    }
                )
                child_processes.append(process)

            # associate scattered processes to process id in yaml
            scattered_processes_path = (
                pathlib.Path(parent_manifest.work_directory)
                / f"scattered_processes_{timestamp_str}.yaml"
            )
            scattered_processes = {
                "parent_pid": os.getpid(),
                "processes": scattered_processes,
            }
            with open(scattered_processes_path, "w") as stream:
                yaml.dump(scattered_processes, stream)
            click.secho(
                f"📊 scattered processes available in {scattered_processes_path}",
                fg="green",
            )

            # ensure all processes completed
            click.secho("🕒 waiting for processes to complete", fg="yellow")
            try:
                for process in child_processes:
                    process.wait()
            except KeyboardInterrupt:
                click.secho(
                    "🚨 caught KeyboardInterrupt, terminating child processes",
                    fg="red",
                )
                for process in child_processes:
                    process.terminate()
                for process in child_processes:
                    process.wait()

            click.secho("✅  all processes completed", fg="green")
        except Exception as exc:
            click.secho(f"{exc}", fg="red")
            _logger.exception(exc)


# TODO: allow user to pass in a particular timestamp?
@cli.command("ps")
@click.pass_context
def ps_cli(ctx):
    """Show status of latest scatter command."""

    try:
        assert "manifest" in ctx.obj, "Manifest not found."
        parent_manifest = ctx.obj["manifest"]

        # get most recent set of scattered manifests
        file_prefix = "scattered_processes_"
        filename_match = f"{file_prefix}*.yaml"

        scattered_processes_path = pathlib.Path(parent_manifest.work_directory)
        scattered_processes_paths = sorted(
            x for x in scattered_processes_path.glob(filename_match)
        )
        if not scattered_processes_paths:
            click.secho(
                f"🚧  no scattered processes found in {parent_manifest.work_directory}/{filename_match}",
                fg="red",
            )
            return
        scattered_processes_path = scattered_processes_paths[-1]

        # list associated info for each process
        state_dir = pathlib.Path(parent_manifest.state_directory)
        with open(scattered_processes_path, "r") as stream:
            scattered_processes = yaml.safe_load(stream)
            for process in scattered_processes["processes"]:
                manifest_path = process["manifest"]
                timestamp_str = (
                    str(manifest_path).split("manifest_scattered_")[1].split(".")[0]
                )

                log_file = "NA"
                metrics_file = "NA"

                try:
                    log_file = list(state_dir.glob(f"vrs_anvil_*{timestamp_str}.log"))[
                        -1
                    ]
                    metrics_file = list(
                        state_dir.glob(f"metrics_*{timestamp_str}.yaml")
                    )[-1]
                except Exception:
                    pass

                click.secho(
                    f"🚧  pid: {str(process['pid'])}, manifest: {str(process['manifest'])}, vcf: {str(process['vcf'])}, metrics_file: {metrics_file}, log_file: {log_file}",
                    fg="yellow",
                )
                process_info = get_process_info(process["pid"])
                if not process_info or metrics_file != "NA":
                    click.secho("  ✅  completed", fg="green")
                else:
                    io_counters = "NA"
                    memory_info = "NA"
                    cpu_percent = "NA"

                    status = process_info.status()
                    if status == "running":
                        try:
                            if hasattr(process_info, "io_counters"):
                                io_counters = process_info.io_counters()
                            if hasattr(process_info, "memory_info"):
                                memory_info = process_info.memory_info()
                            if hasattr(process_info, "cpu_percent"):
                                cpu_percent = process_info.cpu_percent(interval=0.1)
                        except Exception as exc:
                            _logger.info(
                                f"could not get io_counters/memory_info pid: {process['pid']} error:{exc}"
                            )

                        click.secho(
                            f"  📊 {status.capitalize()}: cpu_percent: {cpu_percent}%, io_counters: {io_counters}, memory_info: {memory_info}",
                            fg="yellow",
                        )
    except Exception as exc:
        click.secho(f"{exc}", fg="red")
        _logger.exception(exc)


if __name__ == "__main__":
    cli()



import os
import pytest
import subprocess

from ga4gh.va_spec.base.core import CohortAlleleFrequencyStudyResult as CAF
from ga4gh.vrs.extras.translator import AlleleTranslator
from ga4gh.vrs.dataproxy import create_dataproxy
from plugin_system.plugin_manager import PluginManager
from vrs_anvil.evidence import get_cohort_allele_frequency


############
# FIXTURES #
############


@pytest.fixture()
def existing_vcf_index_path():
    return "tests/fixtures/chr1_multi_sample_index.db"


@pytest.fixture()
def focus_allele_count():
    return 3


@pytest.fixture()
def locus_allele_count():
    return 4


@pytest.fixture()
def phenotype():
    return "HP:0001263"


@pytest.fixture()
def phenotype_index():
    return {
        "sample_A": ["HP:0001263", "HP:0000002"],
        "sample_B": ["HP:0000001"],
    }


@pytest.fixture()
def plugin_methods():
    return [
        "get_phenotype_index",
        "include_sample",
        "process_sample_genotype",
    ]


@pytest.fixture()
def plugin_names():
    return ["BasePlugin", "GregorPlugin", "SimplePlugin"]


@pytest.fixture()
def variant_id():
    return "chr1-20094-TAA-T"


@pytest.fixture()
def vcf_index_path():
    # this path should be empty before
    path = "tests/fixtures/chr1_index.db"

    if os.path.exists(path):
        raise Exception(
            f"VCF index path ({path}) should not exist, please delete before continuing"
        )

    yield path

    if os.path.exists(path):
        os.remove(path)


@pytest.fixture()
def vcf_path():
    return "tests/fixtures/chr1_multi_sample_vrs.vcf.gz"


@pytest.fixture()
def vrs_id():
    return "ga4gh:VA.1WRXw4TC5DYsjC2QLyKux9C0xETLN3Yt"


@pytest.fixture()
def custom_plugin_class_name():
    return "CustomPlugin"


@pytest.fixture()
def custom_plugin_filename():
    return "custom_plugin.py"


@pytest.fixture()
def custom_plugin_dir(custom_plugin_filename, custom_plugin_class_name):
    # write a minimal custom plugin to file
    plugin_class_definition = f"""
from plugin_system.plugins.base_plugin import BasePlugin

class {custom_plugin_class_name}(BasePlugin):
    'stub docstring for {custom_plugin_class_name}'
"""

    with open(custom_plugin_filename, "w") as file:
        file.write(plugin_class_definition)

    # yield enables clean up even on failed pytest
    yield

    # clean up plugin file
    if os.path.exists(custom_plugin_filename):
        os.remove(custom_plugin_filename)


#########
# TESTS #
#########


def test_plugin_manager_can_load_plugins(plugin_methods, plugin_names):

    # load each plugin and ensure it has the proper methods
    for plugin_name in plugin_names:
        print("\n", plugin_name)
        plugin = PluginManager().load_plugin(plugin_name)

        print("plugin:", plugin)
        check_all_plugin_methods_defined(plugin, plugin_name, plugin_methods)


def check_all_plugin_methods_defined(plugin, plugin_name, plugin_methods):
    for method in plugin_methods:
        assert hasattr(plugin, method), f"{plugin_name} does not have method '{method}'"


def test_plugin_manager_loads_default_and_custom_plugin(
    custom_plugin_dir, custom_plugin_class_name, plugin_methods, plugin_names
):
    # NOTE: custom_plugin_dir creates a test file in the working directory instead of a temp dir
    # this is because a temp dir changed at runtime cannot readily be found by the plugin manager's importlib calls
    # could be minor tech debt for devs wishing to run automated tests in the cloud

    # add custom plugin to list of plugin names to load
    all_plugin_names = plugin_names + [custom_plugin_class_name]

    # load all plugins
    for plugin_name in all_plugin_names:
        print(f"\n{plugin_name}")
        plugin = PluginManager().load_plugin(plugin_name)

        check_all_plugin_methods_defined(plugin, plugin_name, plugin_methods)


def test_simple_plugin_can_generate_caf_no_phenotype_index(
    vrs_id, vcf_path, existing_vcf_index_path, focus_allele_count, locus_allele_count
):
    # load plugin class
    plugin_manager = PluginManager()
    plugin = plugin_manager.load_plugin(
        "SimplePlugin"
    )  # SimplePlugin inherits all of base plugin's methods

    # instantiate plugin with no phenotype index
    simple_plugin_no_pheno = plugin()

    # for generated vcf path and vcf_index_path
    caf = get_cohort_allele_frequency(
        variant_id=vrs_id,
        vcf_path=vcf_path,
        vcf_index_path=existing_vcf_index_path,
        plugin=simple_plugin_no_pheno,
    )

    print(f"CAF no phenotype: {caf}")

    # ensure caf is valid and correct
    CAF.model_validate(caf)
    check_allele_counts(caf, vrs_id, focus_allele_count, locus_allele_count)


def test_simple_plugin_can_generate_cafs_with_phenotype_index(
    vrs_id,
    vcf_path,
    existing_vcf_index_path,
    focus_allele_count,
    locus_allele_count,
    phenotype,
    phenotype_index,
):
    # load plugin class
    plugin_manager = PluginManager()
    plugin = plugin_manager.load_plugin(
        "SimplePlugin"
    )  # SimplePlugin inherits all of base plugin's methods

    #####
    # 1 #
    #####
    # create caf using plugin with phenotype index but no phenotype specified
    simple_plugin = plugin(phenotype_index=phenotype_index)
    caf = get_cohort_allele_frequency(
        variant_id=vrs_id,
        vcf_path=vcf_path,
        vcf_index_path=existing_vcf_index_path,
        plugin=simple_plugin,
    )

    print(f"CAF #2: {caf}")

    # check if phenotypes are loaded
    all_phenotypes = [phenotype, "HP:0000002", "HP:0000001"]
    for pheno in all_phenotypes:
        assert (
            pheno in caf.ancillaryResults["phenotypes"]
        ), f'{pheno} not found in caf.ancillary_results["phenotypes"]: {caf.ancillary_results["phenotypes"]}'

    # check if allele counts are accurate
    check_allele_counts(caf, vrs_id, focus_allele_count, locus_allele_count)

    #####
    # 2 #
    #####
    # now, create caf with specified phenotype for subcohort
    caf = get_cohort_allele_frequency(
        variant_id=vrs_id,
        vcf_path=vcf_path,
        vcf_index_path=existing_vcf_index_path,
        plugin=simple_plugin,
        phenotype=phenotype,
    )

    print(f"CAF with phenotype: {caf}")

    # check if phenotypes are loaded
    assert [phenotype] == caf.ancillaryResults[
        "phenotypes"
    ], f"{phenotype} should be the only phenotype in caf.ancillary_results.phenotypes, instead got: {caf.ancillary_results.phenotypes}"

    # check if allele counts are accurate
    check_allele_counts(
        caf=caf, focus_allele=vrs_id, focus_allele_count=1, locus_allele_count=2
    )


def test_plugin_worked_example(
    variant_id,
    vcf_path: str,
    vcf_index_path: str,
    phenotype_index: dict[str, list[str]],
    phenotype: str,
):
    """given a vcf, variant of interest, and phenotype of interest,
    create a

    Args:
        vcf_path (str): path to the VCF of interest
        vcf_index_path (str): path to unwritten index
        phenotype_index (dict[str, list[str]]): mapping from samples to phenotypes
            Index example: {"sample_A": ["HP:0001263", "HP:0000002"], "sample_B": ["HP:0001263"]}
        phenotype (str): phenotype used to specify a subcohort
    """

    # get VRS ID from variant of interest
    seqrepo_rest_service_url = "seqrepo+https://services.genomicmedlab.org/seqrepo"
    seqrepo_dataproxy = create_dataproxy(uri=seqrepo_rest_service_url)
    allele_translator = AlleleTranslator(seqrepo_dataproxy)
    allele = allele_translator.translate_from(variant_id)
    vrs_id = allele.id

    # create vcf index from vcf at the specified path using vrsix
    command = ["vrsix", "load", f"--db-location={vcf_index_path}", vcf_path]
    try:
        result = subprocess.run(command, check=True, text=True, capture_output=True)
        print("Command executed successfully!")
        print("Output:", result.stdout)
    except subprocess.CalledProcessError as e:
        print("Error executing command.")
        print("Error message:", e.stderr)

    # get SimplePlugin class
    plugin_manager = PluginManager()
    plugin = plugin_manager.load_plugin("SimplePlugin")

    # instantiate simple plugin with phenotype index
    simple_plugin = plugin(phenotype_index)

    # generating cohort allele frequency using GREGoR plugin
    caf = get_cohort_allele_frequency(
        variant_id=vrs_id,
        vcf_path=vcf_path,
        vcf_index_path=vcf_index_path,
        plugin=simple_plugin,
        phenotype=phenotype,
    )

    print("caf:", caf)

    check_allele_counts(
        caf=caf, focus_allele=vrs_id, focus_allele_count=1, locus_allele_count=2
    )


# check if allele counts are accurate
def check_allele_counts(
    caf: dict, focus_allele: str, focus_allele_count: int, locus_allele_count: int
) -> None:
    assert (
        caf.focusAllele.root == focus_allele
    ), f"Incorrect CAF: expected focusAllele {focus_allele} but got {caf.focusAllele.root}"

    assert (
        caf.focusAlleleCount == focus_allele_count
    ), f"Incorrect CAF: expected focusAlleleCount {focus_allele_count} but got {caf.focusAlleleCount}"
    assert (
        caf.locusAlleleCount == locus_allele_count
    ), f"Incorrect CAF: expected focusAlleleCount {locus_allele_count} but got {caf.focusAlleleCount}"

    allele_frequency = focus_allele_count * 1.0 / locus_allele_count
    assert (
        caf.focusAlleleFrequency == allele_frequency
    ), f"Incorrect CAF: expected alleleFrequency {allele_frequency} but got {caf.focusAlleleFrequency}"
